"""
====================================================================================================
Spectral Induced Polarization (:mod:`SimPEG.electromagnetics.static.induced_polarization`)
====================================================================================================
.. currentmodule:: SimPEG.electromagnetics.static.spectral_induced_polarization


Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation2DCellCentered
  Simulation2DNodal
  Simulation3DCellCentered
  Simulation3DNodal


Receivers
=========

.. autosummary::
  :toctree: generated/

  receivers.Dipole
  receivers.Pole

Sources
=======
.. autosummary::
  :toctree: generated/

  sources.Dipole
  sources.Pole

Surveys
=======
.. autosummary::
  :toctree: generated/

  survey.Survey

Utilities
=========
.. autosummary::
  :toctree: generated/

  run_inversion
  from_dc_to_sip_survey
  spectral_ip_mappings

Base Classes
============
.. autosummary::
  :toctree: generated/

  receivers.BaseRx
  sources.BaseSrc
  simulation.BaseSIPSimulation
  simulation_2d.BaseSIPSimulation2D

"""
from ....data import Data
from .simulation import Simulation3DCellCentered, Simulation3DNodal
from .simulation_2d import Simulation2DCellCentered, Simulation2DNodal
from .survey import Survey, from_dc_to_sip_survey
from . import sources
from . import receivers
from . import sources as Src
from . import receivers as Rx
from .run import run_inversion, spectral_ip_mappings


Simulation2DCellCentred = Simulation2DCellCentered
Simulation3DCellCentred = Simulation2DCellCentered



"""
==============================================================================
Time-Domain EM (:mod:`SimPEG.electromagnetics.time_domain`)
==============================================================================
.. currentmodule:: SimPEG.electromagnetics.time_domain

About ``time_domain``

Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation1DLayered
  Simulation3DMagneticFluxDensity
  Simulation3DElectricField
  Simulation3DMagneticField
  Simulation3DCurrentDensity

Receivers
=========

.. autosummary::
  :toctree: generated/

  receivers.PointElectricField
  receivers.PointCurrentDensity
  receivers.PointMagneticFluxDensity
  receivers.PointMagneticFluxTimeDerivative
  receivers.PointMagneticField
  receivers.PointMagneticFieldTimeDerivative


Waveforms
=========

.. autosummary::
  :toctree: generated/

  sources.StepOffWaveform
  sources.RampOffWaveform
  sources.RawWaveform
  sources.VTEMWaveform
  sources.TrapezoidWaveform
  sources.TriangularWaveform
  sources.QuarterSineRampOnWaveform
  sources.HalfSineWaveform

Sources
=======

.. autosummary::
  :toctree: generated/

  sources.MagDipole
  sources.CircularLoop
  sources.LineCurrent
  sources.RawVec_Grounded

Surveys
=======
.. autosummary::
  :toctree: generated/

  survey.Survey

Fields
======
.. autosummary::
  :toctree: generated/

  Fields3DMagneticFluxDensity
  Fields3DElectricField
  Fields3DMagneticField
  Fields3DCurrentDensity

Base Classes
============

.. autosummary::
  :toctree: generated/

  receivers.BaseRx
  sources.BaseWaveform
  sources.BaseTDEMSrc
  simulation.BaseTDEMSimulation
  fields.FieldsTDEM
  fields.FieldsDerivativesEB
  fields.FieldsDerivativesHJ

"""
from .simulation import (
    Simulation3DMagneticFluxDensity,
    Simulation3DElectricField,
    Simulation3DMagneticField,
    Simulation3DCurrentDensity,
)
from .simulation_1d import Simulation1DLayered
from .fields import (
    Fields3DMagneticFluxDensity,
    Fields3DElectricField,
    Fields3DMagneticField,
    Fields3DCurrentDensity,
)
from .survey import Survey
from . import sources
from . import receivers

from . import sources as Src
from . import receivers as Rx



import numpy as np

from ....simulation import BaseSimulation
from .... import props

from .survey import Survey

from empymod.transform import get_dlf_points
from empymod import filters
from ....utils import validate_type, validate_string
from scipy.interpolate import InterpolatedUnivariateSpline as iuSpline


HANKEL_FILTERS = [
    "kong_61_2007",
    "kong_241_2007",
    "key_101_2009",
    "key_201_2009",
    "key_401_2009",
    "anderson_801_1982",
    "key_51_2012",
    "key_101_2012",
    "key_201_2012",
    "wer_201_2018",
]


def _phi_tilde(rho, thicknesses, lambdas):
    """Calculate potential in the hankel domain.

    Parameters
    ----------
    rho : (n_layer) np.ndarray
        Resistivity array (ohm m)
    thicknesses : (n_layer - 1) np.ndarray
        Array of layer thicknesses defined from the top down (m).
    lambdas : (n_lambda) np.ndarray
        Wavenumbers.

    Returns
    -------
    (n_lambda) np.ndarray
        Potential in wavenumber domain at sampled locations
    """
    n_layer = len(rho)
    tanh = np.tanh(lambdas[None, :] * thicknesses[:, None])
    t = rho[-1] * np.ones_like(lambdas)
    for i in range(n_layer - 2, -1, -1):
        t = (t + rho[i] * tanh[i]) / (1.0 + t * tanh[i] / rho[i])
    return t


def _dphi_tilde(rho, thicknesses, lambdas):
    """Calculate derivative of potential in the hankel domain.

    Parameters
    ----------
    rho : (n_layer) np.ndarray
        Resistivity array (ohm m)
    thicknesses : (n_layer - 1) np.ndarray
        Array of layer thicknesses defined from the top down (m).
    lambdas : (n_lambda) np.ndarray
        Wavenumbers.

    Returns
    -------
    J_rho : (n_lambda, n_layer) np.ndarray
        Jacobian matrix of first derivatives of lambda w.r.t. resistivity.
    J_h : (n_lambda, n_layer-1) np.ndarray
        Jacobian matrix of first derivatives of lambda w.r.t. thicknesses.
    """
    n_layer = len(rho)
    ts = np.empty((n_layer, len(lambdas)))
    tanh = np.tanh(lambdas[None, :] * thicknesses[:, None])
    tops = np.empty((n_layer, len(lambdas)))
    bots = np.empty((n_layer, len(lambdas)))
    ts[-1] = rho[-1]
    for i in range(n_layer - 2, -1, -1):
        ts[i] = ts[i + 1]
        tops[i] = ts[i + 1] + rho[i] * tanh[i]
        bots[i] = 1 + ts[i + 1] * tanh[i] / rho[i]
        ts[i] = tops[i] / bots[i]
    # return ts[0]
    # ts0 = 1.0
    g_ti = np.ones(len(lambdas))
    J_rho = np.empty((n_layer, len(lambdas)))
    J_h = np.empty((n_layer - 1, len(lambdas)))
    for i in range(n_layer - 1):
        # ts[i] = tops[i] / bots[i]
        g_tops = g_ti / bots[i]
        g_bots = -ts[i] / bots[i] * g_ti
        # bots[i] = 1 + ts[i+1] * tanh[i] / rho0
        g_tip1 = tanh[i] / rho[i] * g_bots
        g_tanh = ts[i + 1] / rho[i] * g_bots
        g_rho0 = -ts[i + 1] * tanh[i] / (rho[i] ** 2) * g_bots
        # tops[i] = ts[i+1] + rho0 * tanh[i]
        g_tip1 += g_tops
        g_tanh += rho[i] * g_tops
        g_rho0 += tanh[i] * g_tops
        # tanh = tanh(thick * lambd)
        g_thick = (1 - tanh[i] ** 2) * lambdas * g_tanh

        J_rho[i] = g_rho0
        J_h[i] = g_thick

        g_ti = g_tip1
    J_rho[-1] = g_ti
    return J_rho.T, J_h.T


class Simulation1DLayers(BaseSimulation):
    """
    1D DC Simulation
    """

    sigma, sigmaMap, sigmaDeriv = props.Invertible("Electrical conductivity (S/m)")
    rho, rhoMap, rhoDeriv = props.Invertible("Electrical resistivity (Ohm m)")
    props.Reciprocal(sigma, rho)

    thicknesses, thicknessesMap, thicknessesDeriv = props.Invertible(
        "thicknesses of the layers"
    )

    def __init__(
        self,
        survey=None,
        sigma=None,
        sigmaMap=None,
        rho=None,
        rhoMap=None,
        thicknesses=None,
        thicknessesMap=None,
        hankel_filter="key_201_2012",
        fix_Jmatrix=False,
        **kwargs,
    ):
        if kwargs.pop("storeJ", None) is not None:
            raise TypeError(
                "storeJ is no longer settable by the user for this simulation."
            )
        if kwargs.pop("hankel_pts_per_dec", None) is not None:
            raise TypeError(
                "hankel_pts_per_dec is no longer settable by the user for this simulation."
            )
        if kwargs.pop("data_type", None) is not None:
            raise TypeError(
                "data_type can no longer be set on the simulation, it must be set on each"
                "receiver."
            )
        super().__init__(survey=survey, **kwargs)
        self.sigma = sigma
        self.rho = rho
        self.thicknesses = thicknesses
        self.sigmaMap = sigmaMap
        self.rhoMap = rhoMap
        self.thicknessesMap = thicknessesMap
        self.fix_Jmatrix = fix_Jmatrix
        self.hankel_filter = hankel_filter  # Store filter
        self._coefficients_set = False
        self._storeJ = True

    @property
    def survey(self):
        """The DC survey object.

        Returns
        -------
        SimPEG.electromagnetics.static.resistivity.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey.")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def storeJ(self):
        """Whether to store the sensitivity matrix.

        Returns
        -------
        bool
        """
        return self._storeJ

    @property
    def hankel_filter(self):
        """The hankel filter key.

        Returns
        -------
        str
        """
        return self._hankel_filter

    @hankel_filter.setter
    def hankel_filter(self, value):
        self._hankel_filter = validate_string(
            "hankel_filter",
            value,
            HANKEL_FILTERS,
        )
        self._fhtfilt = getattr(filters, self._hankel_filter)()

    @property
    def fix_Jmatrix(self):
        """Whether to fix the sensitivity matrix between iterations.

        Returns
        -------
        bool
        """
        return self._fix_Jmatrix

    @fix_Jmatrix.setter
    def fix_Jmatrix(self, value):
        self._fix_Jmatrix = validate_type("fix_Jmatrix", value, bool)

    def _compute_hankel_coefficients(self):
        if self._coefficients_set:
            return
        survey = self.survey

        r_min = np.infty
        r_max = -np.infty

        for src in survey.source_list:
            src_loc = src.location
            for rx in src.receiver_list:
                rx_loc = rx.locations
                if not isinstance(rx_loc, list):
                    # is a pole receiver
                    rx_loc = [rx_loc]
                for loc in rx_loc:
                    off = np.linalg.norm(src_loc[:, None, :] - loc[None, :, :], axis=-1)
                    r_min = min(off.min(), r_min)
                    r_max = max(off.max(), r_max)
        self.survey.set_geometric_factor()

        lambdas, r_spline_points = get_dlf_points(
            self._fhtfilt, np.r_[r_min, r_max], -1
        )
        lambdas = lambdas.reshape(-1)
        n_lambda = len(lambdas)
        n_r = len(r_spline_points)

        n_base = len(self._fhtfilt.base)
        A_dht = np.zeros((n_r, n_lambda))
        for i in range(n_r):
            A_dht[i, i : i + n_base] = self._fhtfilt.j0
        A_dht = A_dht[::-1]  # shuffle these back

        # A_dht goes from wavenumber to space at r_spline_points
        # Then need to spline it from r_spline to all offsets
        # Calculate the interpolating spline basis functions for each spline point
        splines = []
        for i in range(n_r):
            e = np.zeros(n_r)
            e[i] = 1.0
            sp = iuSpline(np.log(r_spline_points[::-1]), e, k=5)
            splines.append(sp)
        # As will go from wavenumber to space domain
        As = []
        for src in survey.source_list:
            src_loc = src.location
            for rx in src.receiver_list:
                rx_loc = rx.locations
                A = np.zeros((rx.nD, n_r))
                for current, tx_elec_loc in zip(src.current, src_loc):
                    if not isinstance(rx_loc, list):
                        # is a pole receiver
                        m_off = np.linalg.norm(tx_elec_loc - rx_loc, axis=-1)
                        n_off = None
                    else:
                        m_off = np.linalg.norm(tx_elec_loc - rx_loc[0], axis=-1)
                        n_off = np.linalg.norm(tx_elec_loc - rx_loc[1], axis=-1)
                    # This receiver has a bunch of data...
                    # this A is the linear operation going from the splined offsets to the data offset
                    for i in range(n_r):
                        A[:, i] += current * splines[i](np.log(m_off)) / m_off
                        if n_off is not None:
                            A[:, i] -= current * splines[i](np.log(n_off)) / n_off
                if rx.data_type == "apparent_resistivity":
                    A /= rx.geometric_factor[src]
                As.append(A @ A_dht / (2 * np.pi))
        self._coefficients_set = True
        self._As = np.vstack(As)
        self._lambdas = lambdas

    def fields(self, m):
        self.model = m
        self._compute_hankel_coefficients()
        return _phi_tilde(self.rho, self.thicknesses, self._lambdas)

    def dpred(self, m=None, f=None):
        """
        Project fields to receiver locations
        :param Fields u: fields object
        :rtype: numpy.ndarray
        :return: data
        """
        if self.verbose:
            print("Calculating predicted data")

        self._compute_hankel_coefficients()
        if f is None:
            if m is None:
                m = self.model
            f = self.fields(m)
        return self._As @ f

    def getJ(self, m, f=None, factor=1e-2):
        """
        Generate Full sensitivity matrix using central difference
        """
        self.model = m
        if getattr(self, "_Jmatrix", None) is None:
            self._compute_hankel_coefficients()
            if self.verbose:
                print("Calculating J and storing")

            J_rho, J_h = _dphi_tilde(self.rho, self.thicknesses, self._lambdas)

            Jmatrix = 0
            if self.rhoMap is not None:
                Jmatrix += (self._As @ J_rho) @ self.rhoDeriv
            if self.thicknessesMap is not None:
                Jmatrix += (self._As @ J_h) @ self.thicknessesDeriv
            self._Jmatrix = Jmatrix
        return self._Jmatrix

    def Jvec(self, m, v, f=None):
        """
        Compute sensitivity matrix (J) and vector (v) product.
        """
        return self.getJ(m, f=f) @ v

    def Jtvec(self, m, v, f=None):
        """
        Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        """
        return self.getJ(m, f=f).T @ v

    @property
    def deleteTheseOnModelUpdate(self):
        to_delete = super().deleteTheseOnModelUpdate
        if not self.fix_Jmatrix:
            to_delete = to_delete + ["_Jmatrix"]
        return to_delete



import numpy as np
from scipy.constants import mu_0

from ...simulation import BaseSimulation
from ... import props
from ...utils import validate_type
from ..frequency_domain.survey import Survey


class Simulation1DRecursive(BaseSimulation):
    r"""
    Simulation class for the 1D MT problem using recursive solution.

    This solution is defined with z +ve upward and a :math:`+i\omega t`
    Fourier convention. First, let:

    .. math::

        \alpha_i^2 = i\omega\mu_i\sigma_i

    The complex impedance in layer :math:`i` is given by:

    .. math::

        Z_i = \dfrac{\alpha_i}{\sigma_i} \Bigg [
        \dfrac{\sigma_i Z_{i+1} - \alpha_i tanh(\alpha_i h_i)}
        {\alpha_i - \sigma_i Z_{i+1}tanh(\alpha_i h_i)} \Bigg ]

    where the complex impedance in the bottom half-space is given by:

    .. math::

        Z_N = - \frac{\alpha_N}{\sigma_N}


    """

    sigma, sigmaMap, sigmaDeriv = props.Invertible("Electrical conductivity (S/m)")
    rho, rhoMap, rhoDeriv = props.Invertible("Electrical resistivity (Ohm m)")
    props.Reciprocal(sigma, rho)

    # Add layer thickness as invertible property
    thicknesses, thicknessesMap, thicknessesDeriv = props.Invertible(
        "thicknesses of the layers starting from the bottom of the mesh"
    )

    def __init__(
        self,
        survey=None,
        sigma=None,
        sigmaMap=None,
        rho=None,
        rhoMap=None,
        thicknesses=None,
        thicknessesMap=None,
        fix_Jmatrix=False,
        **kwargs
    ):
        super().__init__(mesh=None, survey=survey, **kwargs)
        self.fix_Jmatrix = fix_Jmatrix
        self.sigma = sigma
        self.rho = rho
        self.thicknesses = thicknesses
        self.sigmaMap = sigmaMap
        self.rhoMap = rhoMap
        self.thicknessesMap = thicknessesMap

    @property
    def survey(self):
        """The simulations survey.

        Returns
        -------
        SimPEG.electromagnetics.frequency_domain.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey set")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def fix_Jmatrix(self):
        """Whether to fix the sensitivity matrix.

        Returns
        -------
        bool
        """
        return self._fix_Jmatrix

    @fix_Jmatrix.setter
    def fix_Jmatrix(self, value):
        self._fix_Jmatrix = validate_type("fix_Jmatrix", value, bool)

    # TODO: These should be moved to geoana
    def _get_recursive_impedances(self, frequencies, thicknesses, sigmas):
        """
        For a given layered Earth model, this returns the complex impedances
        at the surface for all frequencies.

        Parameters
        ----------
        frequencies : (n_freq, ) np.ndarray
            Frequencies in Hz
        thicknesses : (n_layer-1, ) np.ndarray
            Layer thicknesses in meters, starting from the bottom
        sigmas : (n_layer, ) np.ndarray
            Layer conductivities in S/m, starting from the bottom

        Returns
        -------
        Z : (n_freq, ) np.ndarray
            complex impedances at surface
        """
        frequencies = np.asarray(frequencies)
        thicknesses = np.asarray(thicknesses)[::-1]
        sigmas = np.asarray(sigmas)[::-1]
        omega = 2 * np.pi * frequencies
        n_layer = len(sigmas)

        # layer quantities
        alphas = np.sqrt(1j * omega * mu_0 * sigmas[:, None])
        ratios = alphas / sigmas[:, None]
        tanhs = np.tanh(alphas[:-1] * thicknesses[:, None])

        Z = -ratios[-1]
        # Work from lowest layer to top layer
        for ii in range(n_layer - 2, -1, -1):
            top = Z / ratios[ii] - tanhs[ii]
            bot = 1 - Z / ratios[ii] * tanhs[ii]
            Z = ratios[ii] * top / bot
        return Z

    def _get_recursive_impedances_deriv(self, frequencies, thicknesses, sigmas):
        """
        For a given layered Earth model, this returns the complex impedances
        at the surface for all frequencies.

        Parameters
        ----------
        frequencies : (n_freq, ) np.ndarray
            Frequencies in Hz
        thicknesses : (n_layer-1, ) np.ndarray
            Layer thicknesses in meters, starting from the bottom
        sigmas : (n_layer, ) np.ndarray
            Layer conductivities in S/m, starting from the bottom

        Returns
        -------
        Z : (n_freq, ) np.ndarray
            Complex impedance at surface
        Z_dsigma : (n_freq, n_layer) np.ndarray
            Derivative of complex impedances at surface with respect to sigma
        Z_dsigma : (n_freq, n_layer-1) np.ndarray
            Derivative of complex impedances at surface with respect to thicknesses
        """
        frequencies = np.asarray(frequencies)
        thicknesses = np.asarray(thicknesses)[::-1]
        sigmas = np.asarray(sigmas)[::-1]
        omega = 2 * np.pi * frequencies
        n_layer = len(sigmas)

        # Bottom layer quantities
        alphas = np.sqrt(1j * omega * mu_0 * sigmas[:, None])
        ratios = alphas / sigmas[:, None]
        tanhs = np.tanh(alphas[:-1] * thicknesses[:, None])

        tops = np.empty_like(tanhs)
        bots = np.empty_like(tanhs)
        Zs = np.empty_like(ratios)
        Zs[-1] = -ratios[-1]
        # Work from lowest layer to top layer
        for ii in range(n_layer - 2, -1, -1):
            tops[ii] = Zs[ii + 1] / ratios[ii] - tanhs[ii]
            bots[ii] = 1 - Zs[ii + 1] / ratios[ii] * tanhs[ii]
            Zs[ii] = ratios[ii] * (tops[ii] / bots[ii])

        gZ = 1.0
        gratios = np.empty_like(ratios)
        gtanhs = np.empty_like(tanhs)
        for ii in range(n_layer - 1):
            gratios[ii] = (tops[ii] / bots[ii]) * gZ
            gtop = ratios[ii] / bots[ii] * gZ
            gbot = -Zs[ii] / bots[ii] * gZ

            gZ = -tanhs[ii] / ratios[ii] * gbot
            gratios[ii] += Zs[ii + 1] * tanhs[ii] / ratios[ii] ** 2 * gbot
            gtanhs[ii] = -Zs[ii + 1] / ratios[ii] * gbot

            gZ += gtop / ratios[ii]
            gratios[ii] -= Zs[ii + 1] / ratios[ii] ** 2 * gtop
            gtanhs[ii] -= gtop
        gratios[-1] = -gZ
        d_thick = (1 - tanhs**2) * alphas[:-1] * gtanhs

        galphas = gratios / sigmas[:, None]
        galphas[:-1] += (1 - tanhs**2) * thicknesses[:, None] * gtanhs

        d_sigma = -ratios / sigmas[:, None] * gratios
        d_sigma += (0.5j * omega * mu_0) / alphas * galphas

        # d_mu would be this below when it gets activated:
        # d_mu = (0.5j * omega * sigmas[:, None]) / alphas * galphas
        return Zs[0], d_sigma[::-1].T, d_thick[::-1].T

    def fields(self, m):
        # The layered simulation does not have fields.
        return None

    def dpred(self, m, f=None):
        """
        Computes the data for a given 1D model.

        :param np.array m: inversion model (nP,)
        :return np.array f: data (nD,)
        """
        self.model = m

        # Compute complex impedances for each frequency=
        Z = self._get_recursive_impedances(
            self.survey.frequencies, self.thicknesses, self.sigma
        )

        # For each complex impedance, extract compute datum
        d = []
        for src in self.survey.source_list:
            i_freq = np.searchsorted(self.survey.frequencies, src.frequency)
            for rx in src.receiver_list:
                if rx.component == "real":
                    d.append(np.real(Z[i_freq]))
                elif rx.component == "imag":
                    d.append(np.imag(Z[i_freq]))
                elif rx.component == "apparent_resistivity":
                    d.append(
                        np.abs(Z[i_freq]) ** 2 / (2 * np.pi * src.frequency * mu_0)
                    )
                elif rx.component == "phase":
                    d.append(
                        (180.0 / np.pi)
                        * np.arctan(np.imag(Z[i_freq]) / np.real(Z[i_freq]))
                    )

        return np.array(d)

    def getJ(self, m, f=None):
        """
        Compute and store the sensitivity matrix.

        :param numpy.ndarray m: inversion model (nP,)
        :return numpy.ndarray J: Sensitivity matrix (nD, nP)
        """
        # Analytic computation
        self.model = m
        if getattr(self, "_Jmatrix", None) is not None:
            return self._Jmatrix

        # Derivatives for conductivity
        Z, Z_dsigma, Z_dthick = self._get_recursive_impedances_deriv(
            self.survey.frequencies, self.thicknesses, self.sigma
        )
        Js = []
        if self.sigmaMap is not None:
            Js.append(Z_dsigma)
        if self.thicknessesMap is not None:
            Js.append(Z_dthick)
        Js = np.hstack(Js)

        J = np.empty((self.survey.nD, Js.shape[1]))

        start = 0
        for src in self.survey.source_list:
            i_freq = np.searchsorted(self.survey.frequencies, src.frequency)
            Js_row = Js[i_freq]
            for rx in src.receiver_list:
                if rx.component == "real":
                    Jrows = np.real(Js_row)
                elif rx.component == "imag":
                    Jrows = np.imag(Js_row)
                elif rx.component == "apparent_resistivity":
                    Jrows = (np.pi * src.frequency * mu_0) ** -1 * (
                        np.real(Z[i_freq]) * np.real(Js_row)
                        + np.imag(Z[i_freq]) * np.imag(Js_row)
                    )
                elif rx.component == "phase":
                    C = 180 / np.pi
                    real = np.real(Z[i_freq])
                    imag = np.imag(Z[i_freq])
                    bot = real**2 + imag**2
                    d_real_dm = np.real(Js_row)
                    d_imag_dm = np.imag(Js_row)
                    Jrows = C * (-imag / bot * d_real_dm + real / bot * d_imag_dm)
                end = start + rx.nD
                J[start:end] = Jrows
                start = end
        self._Jmatrix = {}
        start = 0
        if self.sigmaMap is not None:
            end = start + Z_dsigma.shape[1]
            self._Jmatrix["sigma"] = J[:, start:end]
            start = end
        if self.thicknessesMap is not None:
            end = start + Z_dthick.shape[1]
            self._Jmatrix["thick"] = J[:, start:end]
        return self._Jmatrix

    def getJtJdiag(self, m, W=None, f=None):
        if getattr(self, "_jtjdiag", None) is None:
            Js = self.getJ(m, f=f)
            if W is None:
                W = np.ones(self.survey.nD)
            else:
                W = W.diagonal() ** 2

            jtjdiag = 0
            if self.sigmaMap is not None:
                J = Js["sigma"] @ self.sigmaDeriv
                jtjdiag += np.einsum("i,ij,ij->j", W, J, J)
            if self.thicknessesMap is not None:
                J = Js["thick"] @ self.thicknessesDeriv
                jtjdiag += np.einsum("i,ij,ij->j", W, J, J)
            self._jtjdiag = jtjdiag
        return self._jtjdiag

    def Jvec(self, m, v, f=None):
        J = self.getJ(m, f=None)
        Jvec = 0
        if self.sigmaMap is not None:
            Jvec += J["sigma"] @ (self.sigmaDeriv * v)
        if self.thicknessesMap is not None:
            Jvec += J["thick"] @ (self.thicknessesDeriv * v)
        return Jvec

    def Jtvec(self, m, v, f=None):
        J = self.getJ(m, f=None)
        JTvec = 0
        if self.sigmaMap is not None:
            JTvec += self.sigmaDeriv.T @ (J["sigma"].T @ v)
        if self.thicknessesMap is not None:
            JTvec += self.thicknessesDeriv.T @ (J["thick"].T @ v)
        return JTvec

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super().deleteTheseOnModelUpdate
        if self.fix_Jmatrix:
            return toDelete
        else:
            toDelete = toDelete + ["_Jmatrix", "_jtjdiag"]
        return toDelete



import numpy as np
import scipy.special as spec
from ...utils import (
    validate_string,
    validate_float,
    validate_ndarray_with_shape,
    validate_callable,
)


class BaseVRMWaveform:
    """
    The Base VRM Waveform object
    """

    pass
    # included to have an identifiable base class for the VRM waveforms


###################################################
#           STEP OFF WAVEFORM
###################################################


class StepOff(BaseVRMWaveform):
    """Characteristic decay class for step-off waveform

    Parameters
    ----------
    t0 : float
        Beginning of the off-time
    """

    def __init__(self, t0=0.0):
        self.t0 = t0

    @property
    def t0(self):
        """Beginning of the off-time

        Returns
        -------
        float
            Beginning of the off-time
        """
        return self._t0

    @t0.setter
    def t0(self, value):
        self._t0 = validate_float("t0", value)

    def getCharDecay(self, fieldType, times):
        """Return characteristic decay for step-off waveform.

        This function defines the decay of the VRM response for the linear problem type.
        Note that the current will be normalized by its maximum value. The
        maximum current in the transmitter is specified in the source object.

        Parameters
        ----------
        fieldType : {'dhdt', 'dbdt'}
            Field type. Characteristic decay for 'h'
            or 'b' CANNOT be computed for step-off
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["dhdt", "dbdt"])

        if self.t0 >= np.min(times):
            raise ValueError(
                "Earliest time channel must be after beginning of off-time (t0 = %.2e s)"
                % self.t0
            )

        t0 = self.t0

        if fieldType == "dbdt":
            mu0 = 4 * np.pi * 1e-7
            eta = -mu0 / (times - t0)
        elif fieldType == "dhdt":
            eta = -1 / (times - t0)

        return eta

    def getLogUniformDecay(self, fieldType, times, chi0, dchi, tau1, tau2):
        """Return characteristic decay for a step-off waveform for a log-uniform distribution of time-relaxation constants.

        The output of this function is the magnetization at each time for each cell, normalized by the inducing
        field.

        Parameters
        ----------
        fieldType : {'dhdt', 'dbdt'}
            Field type.
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.
        chi0 : float
            DC (zero-frequency) magnetic susceptibility for all cells
        dchi : float
            DC (zero-frequency) magnetic susceptibility attributed to VRM for all cells
        tau1 : float
            Lower-bound for log-uniform distribution of time-relaxation
            constants for all cells
        tau2 : float
            Upper-bound for log-uniform distribution of time-relaxation
            constants for all cells

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["dhdt", "dbdt"])

        nT = len(times)
        nC = len(dchi)
        t0 = self.t0

        times = np.kron(np.ones((nC, 1)), times)
        chi0 = np.kron(np.reshape(chi0, newshape=(nC, 1)), np.ones((1, nT)))
        dchi = np.kron(np.reshape(dchi, newshape=(nC, 1)), np.ones((1, nT)))
        tau1 = np.kron(np.reshape(tau1, newshape=(nC, 1)), np.ones((1, nT)))
        tau2 = np.kron(np.reshape(tau2, newshape=(nC, 1)), np.ones((1, nT)))

        if fieldType == "h":
            eta = 0.5 * (1 - np.sign(times - t0)) * chi0 + 0.5 * (
                1 + np.sign(times - t0)
            ) * (dchi / np.log(tau2 / tau1)) * (
                spec.expi(-(times - t0) / tau2) - spec.expi(-(times - t0) / tau1)
            )
        elif fieldType == "b":
            mu0 = 4 * np.pi * 1e-7
            eta = 0.5 * (1 - np.sign(times - t0)) * chi0 + 0.5 * (
                1 + np.sign(times - t0)
            ) * (dchi / np.log(tau2 / tau1)) * (
                spec.expi(-(times - t0) / tau2) - spec.expi(-(times - t0) / tau1)
            )
            eta = mu0 * eta
        elif fieldType == "dhdt":
            eta = 0.0 + 0.5 * (1 + np.sign(times - t0)) * (
                dchi / np.log(tau2 / tau1)
            ) * (np.exp(-(times - t0) / tau1) - np.exp(-(times - t0) / tau2)) / (
                times - t0
            )
        elif fieldType == "dbdt":
            mu0 = 4 * np.pi * 1e-7
            eta = 0.0 + 0.5 * (1 + np.sign(times - t0)) * (
                dchi / np.log(tau2 / tau1)
            ) * (np.exp(-(times - t0) / tau1) - np.exp(-(times - t0) / tau2)) / (
                times - t0
            )
            eta = mu0 * eta

        return eta


###################################################
#           SQUARE PULSE WAVEFORM
###################################################


class SquarePulse(StepOff):
    """Characteristic decay class for square-pulse waveform

    Parameters
    ----------
    delt : float
        Pulse width
    t0 : float
        Beginning of the off-time
    """

    def __init__(self, delt, t0=0.0):
        super(SquarePulse, self).__init__(t0=t0)

        self.delt = delt

    @property
    def delt(self):
        """Square pulse on-time length

        Returns
        -------
        float
            Square pulse on-time length
        """
        return self._delt

    @delt.setter
    def delt(self, value):
        self._delt = validate_float("delt", value, min_val=0.0, inclusive_min=False)

    def getCharDecay(self, fieldType, times):
        """Compute characteristic decay for a square-pulse waveform.

        This function describes the decay of the VRM response for the linear
        problem type. Note that the current will be normalized by its maximum
        value. The maximum current in the transmitter is specified in the
        source object.

        Parameters
        ----------
        fieldType : {'h', 'b', 'dhdt', 'dbdt'}
            Field type.
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["h", "b", "dhdt", "dbdt"])

        if self.t0 >= np.min(times):
            raise ValueError(
                "Earliest time channel must be after beginning of off-time (t0 = %.2e s)"
                % self.t0
            )

        t0 = self.t0
        delt = self.delt
        mu0 = 4 * np.pi * 1e-7

        if fieldType == "h":
            eta = np.log(1 + delt / (times - t0))
        elif fieldType == "b":
            eta = mu0 * np.log(1 + delt / (times - t0))
        elif fieldType == "dhdt":
            eta = -(1 / (times - t0) - 1 / (times - t0 + delt))
        elif fieldType == "dbdt":
            eta = -mu0 * (1 / (times - t0) - 1 / (times - t0 + delt))

        return eta

    def getLogUniformDecay(self, fieldType, times, chi0, dchi, tau1, tau2):
        """Characteristic decay for a square-pulse waveform for log-uniform distribution of time-relaxation constants.

        The output of this function is the magnetization at each time for each cell, normalized by the inducing field.

        The output of this function is the magnetization at each time for each cell, normalized by the inducing
        field.

        Parameters
        ----------
        fieldType : {'h', 'b', 'dhdt', 'dbdt'}
            Field type.
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.
        chi0 : float
            DC (zero-frequency) magnetic susceptibility for all cells
        dchi : float
            DC (zero-frequency) magnetic susceptibility attributed to VRM for all cells
        tau1 : float
            Lower-bound for log-uniform distribution of time-relaxation
            constants for all cells
        tau2 : float
            Upper-bound for log-uniform distribution of time-relaxation
            constants for all cells

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["h", "b", "dhdt", "dbdt"])

        nT = len(times)
        nC = len(dchi)
        t0 = self.t0
        delt = self.delt

        times = np.kron(np.ones((nC, 1)), times)
        chi0 = np.kron(np.reshape(chi0, newshape=(nC, 1)), np.ones((1, nT)))
        dchi = np.kron(np.reshape(dchi, newshape=(nC, 1)), np.ones((1, nT)))
        tau1 = np.kron(np.reshape(tau1, newshape=(nC, 1)), np.ones((1, nT)))
        tau2 = np.kron(np.reshape(tau2, newshape=(nC, 1)), np.ones((1, nT)))

        if fieldType == "h":
            eta = (np.sign(times - t0 + delt) - np.sign(times - t0)) * (
                chi0 - dchi
            ) - 0.5 * (1 + np.sign(times - t0)) * (dchi / np.log(tau2 / tau1)) * (
                spec.expi(-(times - t0) / tau2)
                - spec.expi(-(times - t0) / tau1)
                - spec.expi(-(times - t0 + delt) / tau2)
                + spec.expi(-(times - t0 + delt) / tau1)
            )
        elif fieldType == "b":
            mu0 = 4 * np.pi * 1e-7
            eta = (np.sign(times - t0 + delt) - np.sign(times - t0)) * (
                chi0 - dchi
            ) - 0.5 * (1 + np.sign(times - t0)) * (dchi / np.log(tau2 / tau1)) * (
                spec.expi(-(times - t0) / tau2)
                - spec.expi(-(times - t0) / tau1)
                - spec.expi(-(times - t0 + delt) / tau2)
                + spec.expi(-(times - t0 + delt) / tau1)
            )
            eta = mu0 * eta
        elif fieldType == "dhdt":
            eta = (
                0.0
                + 0.5
                * (1 + np.sign(times - t0))
                * (dchi / np.log(tau2 / tau1))
                * (np.exp(-(times - t0) / tau1) - np.exp(-(times - t0) / tau2))
                / (times - t0)
                - 0.5
                * (1 + np.sign(times - t0 + delt))
                * (dchi / np.log(tau2 / tau1))
                * (
                    np.exp(-(times - t0 + delt) / tau1)
                    - np.exp(-(times - t0 + delt) / tau2)
                )
                / (times - t0 + delt)
            )
        elif fieldType == "dbdt":
            mu0 = 4 * np.pi * 1e-7
            eta = (
                0.0
                + 0.5
                * (1 + np.sign(times - t0))
                * (dchi / np.log(tau2 / tau1))
                * (np.exp(-(times - t0) / tau1) - np.exp(-(times - t0) / tau2))
                / (times - t0)
                - 0.5
                * (1 + np.sign(times - t0 + delt))
                * (dchi / np.log(tau2 / tau1))
                * (
                    np.exp(-(times - t0 + delt) / tau1)
                    - np.exp(-(times - t0 + delt) / tau2)
                )
                / (times - t0 + delt)
            )
            eta = mu0 * eta

        return eta


###################################################
#    ARBITRARY WAVEFORM UNIFORM DISCRETIZATION
###################################################


class ArbitraryDiscrete(BaseVRMWaveform):
    """Characteristic decay for arbitrary discrete waveform

    This class is used to approximate an arbitrary waveform as a set of square-pulse waveforms;
    for which we have a solution to the characteristic decay.

    Parameters
    ----------
    t_wave : numpy.ndarray
        Waveform on-times
    I_wave : numpy.ndarray
        Waveform on-time currents
    """

    def __init__(self, t_wave, I_wave):
        self.t_wave = t_wave
        self.I_wave = I_wave

    @property
    def t_wave(self):
        """Waveform on-times

        Returns
        -------
        numpy.ndarray
            Waveform on-times
        """
        return self._t_wave

    @t_wave.setter
    def t_wave(self, value):
        value = validate_ndarray_with_shape("t_wave", value, shape=("*",))
        if getattr(self, "_I_wave", None) is not None and len(value) != len(
            self._I_wave
        ):
            raise ValueError("'t_wave' and 'I_wave' must be the same length")
        self._t_wave = value

    @property
    def I_wave(self):
        """Waveform on-time currents

        Returns
        -------
        numpy.ndarray
            Waveform on-time currents
        """
        return self._I_wave

    @I_wave.setter
    def I_wave(self, value):
        value = validate_ndarray_with_shape("I_wave", value, shape=("*",))
        if getattr(self, "_t_wave", None) is not None and len(value) != len(
            self._t_wave
        ):
            raise ValueError("'t_wave' and 'I_wave' must be the same length")
        self._I_wave = value

    def getCharDecay(self, fieldType, times):
        """Compute characteristic decay for arbitrary waveform.

        This function describes the decay of the VRM response for the Linear problem type.
        Note that the current will be normalized by its maximum value. The
        maximum current in the transmitter is specified in the source object.

        Parameters
        ----------
        fieldType : {'h', 'b', 'dhdt', 'dbdt'}
            Field type.
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["h", "b", "dhdt", "dbdt"])

        k = np.where(self.I_wave > 1e-10)
        j = k[0][0] - 1
        k = k[0][-1] + 1

        twave = self.t_wave[j : k + 1]
        Iwave = self.I_wave[j : k + 1] / np.max(np.abs(self.I_wave[j : k + 1]))

        n_pts = int(np.ceil(25 * (np.max(twave) - np.min(twave)) / np.min(times)))

        if n_pts > 25000:
            n_pts = 25000

        dt = (np.max(twave) - np.min(twave)) / np.float64(n_pts)
        tvec = np.linspace(np.min(twave), np.max(twave), n_pts + 1)

        g = np.r_[Iwave[0], np.interp(tvec[1:-1], twave, Iwave), Iwave[-1]]
        tvec = tvec[1:]

        eta = np.zeros(len(times))

        if fieldType in ["h", "b"]:
            for tt in range(0, len(eta)):
                eta[tt] = np.sum(
                    (g[1:] + (g[1:] - g[0:-1]) * (times[tt] - tvec) / dt)
                    * np.log(1 + dt / (times[tt] - tvec))
                    - g[1:]
                    + g[0:-1]
                )
        elif fieldType in ["dhdt", "dbdt"]:
            for tt in range(0, len(eta)):
                eta[tt] = np.sum(
                    ((g[1:] - g[0:-1]) / dt) * np.log(1 + dt / (times[tt] - tvec))
                    - (g[1:] + (g[1:] - g[0:-1]) * (times[tt] - tvec) / dt)
                    * (1 / (times[tt] - tvec + dt) - 1 / (times[tt] - tvec))
                )

        if fieldType in ["b", "dbdt"]:
            mu0 = 4 * np.pi * 1e-7
            eta = mu0 * eta

        return eta


###################################################
#    ARBITRARY WAVEFORM PIECEWISE DISCRETIZATION
###################################################


class ArbitraryPiecewise(ArbitraryDiscrete):
    """Characteristic decay for arbitrary piecewise waveform

    This class is used to approximate an arbitrary waveform using a piecewise linear approximation;
    for which we have a solution to the characteristic decay.

    Parameters
    ----------
    t_wave : numpy.ndarray
        Waveform on-times
    I_wave : numpy.ndarray
        Waveform on-time currents
    """

    def __init__(self, t_wave, I_wave):
        super(ArbitraryPiecewise, self).__init__(t_wave=t_wave, I_wave=I_wave)

    def getCharDecay(self, fieldType, times):
        """Compute characteristic decay function for arbitrary waveform.

        This function describes the decay of the VRM response for the Linear problem type.
        Note that the current will be LogUniformized by its maximum value. The
        maximum current in the transmitter is specified in the source object.

        Parameters
        ----------
        fieldType : {'h', 'b', 'dhdt', 'dbdt'}
            Field type.
        times : numpy.ndarray
            Observation times. These times MUST be during the off-time.

        Returns
        -------
        eta : (n_times) numpy.ndarray
            Characteristic decay evaluated at all specified times.
        """
        fieldType = validate_string("fieldType", fieldType, ["h", "b", "dhdt", "dbdt"])

        if np.max(self.t_wave) >= np.min(times):
            raise ValueError(
                "Earliest time channel must be after beginning of off-time (t0 = %.2e s)"
                % np.max(self.t_wave)
            )

        k = np.where(self.I_wave > 1e-10)
        j = k[0][0] - 1
        k = k[0][-1] + 1

        tvec = self.t_wave[j : k + 1]
        dt = tvec[1:] - tvec[0:-1]
        g = self.I_wave[j : k + 1] / np.max(np.abs(self.I_wave[j : k + 1]))
        tvec = tvec[1:]

        eta = np.zeros(len(times))

        if fieldType in ["h", "b"]:
            for tt in range(0, len(eta)):
                eta[tt] = np.sum(
                    (g[1:] + (g[1:] - g[0:-1]) * (times[tt] - tvec) / dt)
                    * np.log(1 + dt / (times[tt] - tvec))
                    - g[1:]
                    + g[0:-1]
                )
        elif fieldType in ["dhdt", "dbdt"]:
            for tt in range(0, len(eta)):
                eta[tt] = np.sum(
                    ((g[1:] - g[0:-1]) / dt) * np.log(1 + dt / (times[tt] - tvec))
                    - (g[1:] + (g[1:] - g[0:-1]) * (times[tt] - tvec) / dt)
                    * (1 / (times[tt] - tvec + dt) - 1 / (times[tt] - tvec))
                )

        if fieldType in ["b", "dbdt"]:
            mu0 = 4 * np.pi * 1e-7
            eta = mu0 * eta

        return eta


###################################################
#               CUSTOM DECAY
###################################################


class Custom(BaseVRMWaveform):
    """Define characteristic decay with function handle

    Parameters
    ----------
    waveform_function : function
        Function handle defining the characteristic decay as a function of time
    """

    def __init__(self, waveform_function):
        self.waveform_function = waveform_function

    @property
    def waveform_function(self):
        """Function handle for characteristic decay

        Returns
        -------
        function
            A function that returns the characteristic decay as a
            function of *times*.
        """
        return self._waveform_function

    @waveform_function.setter
    def waveform_function(self, value):
        self._waveform_function = validate_callable("waveform_function", value)

    def getCharDecay(self, times):
        """Returns characteristic decay function at specified times

        Parameters
        ----------
        times : numpy.ndarray
            Off-times where characteristic decay is evaluated
        """

        return self.waveform_function(times)



import numpy as np
from .code_utils import deprecate_function
from discretize.utils import (  # noqa: F401
    Zero,
    Identity,
    mkvc,
    sdiag,
    sdinv,
    speye,
    kron3,
    spzeros,
    ddx,
    av,
    av_extrap,
    ndgrid,
    ind2sub,
    sub2ind,
    get_subarray,
    inverse_3x3_block_diagonal,
    inverse_2x2_block_diagonal,
    TensorType,
    make_property_tensor,
    inverse_property_tensor,
)


def estimate_diagonal(matrix_arg, n, k=None, approach="Probing"):
    r"""Estimate the diagonal of a matrix.

    This function estimates the diagonal of a matrix using one of the following
    iterative methods:

        - **Probing:** cyclic permutations of vectors with 1's and 0's (default)
        - **Ones:** random +/- 1 entries
        - **Random:** random vectors with entries in the range [-1, 1]

    The user can estimate the diagonal of the matrix by providing the matrix,
    or by providing a function hangle which computes the dot product of the matrix
    and a vector.

    For background information on this method, see
    `Bekas (et al., 2005) <http://www-users.cs.umn.edu/~saad/PDF/umsi-2005-082.pdf>`__
    and `Selig (et al., 2012) <https://www.cita.utoronto.ca/~niels/diagonal.pdf>`__

    Parameters
    ----------
    matrix_arg : numpy.ndarray or function
        The matrix as a ``numpy.ndarray``, or a function handle which computes the dot product
        between the matrix and a vector.
    n : int
        The length of the random vectors used to compute the diagonal; equals number of columns
    k : int
        Number of vectors to be used to estimate the diagonal; i.e. number of iterations in estimation
    approach : str
        Method used for approximating diagonal. Must be one of {'probing', 'ones', 'random'}

    Returns
    -------
    numpy.ndarray
        Estimate of the diagonal elements of the input matrix

    """

    if isinstance(matrix_arg, np.ndarray):
        A = matrix_arg

        def matrix_arg(v):
            return A.dot(v)

    if k is None:
        k = np.floor(n / 10.0)

    if approach.upper() == "ONES":

        def getv(n, i=None):
            v = np.random.randn(n)
            v[v < 0] = -1.0
            v[v >= 0] = 1.0
            return v

    elif approach.upper() == "RANDOM":

        def getv(n, i=None):
            return np.random.randn(n)

    else:  # if approach == 'Probing':

        def getv(n, i):
            v = np.zeros(n)
            v[i:n:k] = 1.0
            return v

    Mv = np.zeros(n)
    vv = np.zeros(n)

    for i in range(0, k):
        vk = getv(n, i)
        Mv += matrix_arg(vk) * vk
        vv += vk * vk

    d = Mv / vv

    return d


def unique_rows(M):
    """Return unique rows, row indices and inverse indices.

    Parameters
    ----------
    M : array_like
        The input array

    Returns
    -------
    unqM : numpy.ndarray
        Array consisting of the unique rows of the input array
    unqInd : numpy.ndarray of int
        Indices to project from input array to output array
    unqInd : numpy.ndarray of int
        Indices to project from output array to input array

    """
    b = np.ascontiguousarray(M).view(np.dtype((np.void, M.dtype.itemsize * M.shape[1])))
    _, unqInd = np.unique(b, return_index=True)
    _, invInd = np.unique(b, return_inverse=True)
    unqM = M[unqInd]
    return unqM, unqInd, invInd


def eigenvalue_by_power_iteration(
    combo_objfct, model, n_pw_iter=4, fields_list=None, seed=None
):
    r"""Estimate largest eigenvalue in absolute value using power iteration.

    Uses the power iteration approach to estimate the largest eigenvalue in absolute
    value for a single :class:`SimPEG.BaseObjectiveFunction` or a combination of
    objective functions stored in a :class:`SimPEG.ComboObjectiveFunction`.

    Parameters
    ----------
    combo_objfct : SimPEG.BaseObjectiveFunction
        Objective function or a combo objective function
    model : numpy.ndarray
        Current model
    n_pw_iter : int
        Number of power iterations used to estimate the highest eigenvalue
    fields_list : list (optional)
        ``list`` of fields objects for each data misfit term in combo_objfct. If none given,
        they will be evaluated within the function. If combo_objfct mixs data misfit and regularization
        terms, the list should contains SimPEG.fields for the data misfit terms and None for the
        regularization term.
    seed : int
        Random seed for the initial random guess of eigenvector.

    Returns
    -------
    float
        Estimated value of the highest eigenvalue in absolute value

    Notes
    -----
    After *k* power iterations, the largest eigenvalue in absolute value is
    approximated by the Rayleigh quotient:

    .. math::
        \lambda_k = \frac{\mathbf{x_k^T A x_k}}{\mathbf{x_k^T x_k}}

    where :math:`\mathfb{A}` is our matrix and :math:`\mathfb{x_k}` is computed
    recursively according to:

    .. math::
        \mathbf{x_{k+1}} = \frac{\mathbf{A x_k}}{\| \mathbf{Ax_k} \|}

    The elements of the initial vector :math:`\mathbf{x_0}` are randomly
    selected from a uniform distribution.

    """

    if seed is not None:
        np.random.seed(seed)

    # Initial guess for eigen-vector
    x0 = np.random.rand(*model.shape)
    x0 = x0 / np.linalg.norm(x0)

    # transform to ComboObjectiveFunction if required
    if getattr(combo_objfct, "objfcts", None) is None:
        combo_objfct = 1.0 * combo_objfct

    # create Field for data misfit if necessary and not provided
    if fields_list is None:
        fields_list = []
        for obj in combo_objfct.objfcts:
            if hasattr(obj, "simulation"):
                fields_list += [obj.simulation.fields(model)]
            else:
                # required to put None to conserve it in the list
                # The idea is that the function can have a mixed of dmis and reg terms
                # (see test)
                fields_list += [None]
    elif not isinstance(fields_list, (list, tuple, np.ndarray)):
        fields_list = [fields_list]

    # Power iteration: estimate eigenvector
    for _ in range(n_pw_iter):
        x1 = 0.0
        for j, (mult, obj) in enumerate(
            zip(combo_objfct.multipliers, combo_objfct.objfcts)
        ):
            if hasattr(obj, "simulation"):  # if data misfit term
                aux = obj.deriv2(model, v=x0, f=fields_list[j])
                if not isinstance(aux, Zero):
                    x1 += mult * aux
            else:
                aux = obj.deriv2(model, v=x0)
                if not isinstance(aux, Zero):
                    x1 += mult * aux
        x0 = x1 / np.linalg.norm(x1)

    # Compute highest eigenvalue from estimated eigenvector
    eigenvalue = 0.0
    for j, (mult, obj) in enumerate(
        zip(combo_objfct.multipliers, combo_objfct.objfcts)
    ):
        if hasattr(obj, "simulation"):  # if data misfit term
            eigenvalue += mult * x0.dot(obj.deriv2(model, v=x0, f=fields_list[j]))
        else:
            eigenvalue += mult * x0.dot(
                obj.deriv2(
                    model,
                    v=x0,
                )
            )

    return eigenvalue


def cartesian2spherical(m):
    r"""
    Converts a set of 3D vectors from Cartesian to spherical coordinates.

    Parameters
    ----------
    m : (n, 3) array_like
        An array whose columns represent the x, y and z components of
        a set of vectors.

    Returns
    -------
    (n, 3) numpy.ndarray
        An array whose columns represent the *a*, *t* and *p* components
        of a set of vectors in spherical coordinates.

    Notes
    -----

    In Cartesian space, the components of each vector are defined as

    .. math::

        \mathbf{v} = (v_x, v_y, v_z)

    In spherical coordinates, vectors are is defined as:

    .. math::

        \mathbf{v^\prime} = (a, t, p)

    where

        - :math:`a` is the amplitude of the vector
        - :math:`t` is the azimuthal angle defined positive from vertical
        - :math:`p` is the radial angle defined positive CCW from Easting

    """

    # nC = int(len(m)/3)

    x = m[:, 0]
    y = m[:, 1]
    z = m[:, 2]

    a = (x**2.0 + y**2.0 + z**2.0) ** 0.5

    t = np.zeros_like(x)
    t[a > 0] = np.arcsin(z[a > 0] / a[a > 0])

    p = np.zeros_like(x)
    p[a > 0] = np.arctan2(y[a > 0], x[a > 0])

    m_atp = np.r_[a, t, p]

    return m_atp


def cartesian2amplitude_dip_azimuth(m):
    """
    Convert from cartesian to amplitude, dip (positive down) and
    azimuth (clockwise for North), in degree.
    """
    m = m.reshape((-1, 3), order="F")
    atp = cartesian2spherical(m).reshape((-1, 3), order="F")
    atp[:, 1] = np.rad2deg(-1.0 * atp[:, 1])
    atp[:, 2] = (450.0 - np.rad2deg(atp[:, 2])) % 360.0

    return atp


def spherical2cartesian(m):
    """Convert from spherical to cartesian"""
    m = m.reshape((-1, 3), order="F")
    a = m[:, 0] + 1e-8
    t = m[:, 1]
    p = m[:, 2]
    m_xyz = np.r_[a * np.cos(t) * np.cos(p), a * np.cos(t) * np.sin(p), a * np.sin(t)]

    return m_xyz


def spherical2cartesian(m):
    r"""
    Converts a set of 3D vectors from spherical to Catesian coordinates.

    Parameters
    ----------
    m : (n, 3) array_like
        An array whose columns represent the *a*, *t* and *p* components of
        a set of vectors in spherical coordinates.

    Returns
    -------
    (n, 3) numpy.ndarray
        An array whose columns represent the *x*, *y* and *z* components
        of the set of vectors in Cartesian.

    Notes
    -----

    In Cartesian space, the components of each vector are defined as

    .. math::

        \mathbf{v} = (v_x, v_y, v_z)

    In spherical coordinates, vectors are is defined as:

    .. math::

        \mathbf{v^\prime} = (a, t, p)

    where

        - :math:`a` is the amplitude of the vector
        - :math:`t` is the azimuthal angle defined positive from vertical
        - :math:`p` is the radial angle defined positive CCW from Easting

    """
    m = m.reshape((-1, 3), order="F")
    a = m[:, 0] + 1e-8
    t = m[:, 1]
    p = m[:, 2]

    m_xyz = np.r_[a * np.cos(t) * np.cos(p), a * np.cos(t) * np.sin(p), a * np.sin(t)]

    return m_xyz


def dip_azimuth2cartesian(dip, azm):
    """Convert vectors from dip-azimuth to Cartesian

    This function takes the dip and azimuthal angles for a set of vectors and
    converts to Cartesian coordinates. The output is a numpy.ndarray whose
    columns represent the vectors' x, y and z components.

    Parameters
    ----------
    dip : float or 1D numpy.ndarray
        Dip angle in degrees. Values in range [0, 90]
    azm : float or 1D numpy.ndarray
        Asimuthal angle (strike) in degrees. Defined clockwise from Northing. Values is range [0, 360]

    Returns
    -------
    (n, 3) numpy.ndarray
        Numpy array whose columns represent the x, y and z components of the
        vector(s) in Cartesian coordinates

    """

    azm = np.asarray(azm)
    dip = np.asarray(dip)

    # Number of elements
    nC = azm.size

    M = np.zeros((nC, 3))

    # Modify azimuth from North to cartesian-X
    azm_X = (450.0 - np.asarray(azm)) % 360.0
    inc = -np.deg2rad(np.asarray(dip))
    dec = np.deg2rad(azm_X)

    M[:, 0] = np.cos(inc) * np.cos(dec)
    M[:, 1] = np.cos(inc) * np.sin(dec)
    M[:, 2] = np.sin(inc)

    return M


def coterminal(theta):
    r"""
    Compute coterminal angle

    For a set of angles defined in radians, this function outputs their coterminal angles.
    That is, for an angle :math:`\theta` where:

    .. math::

        \theta = 2\pi N + \gamma

    and *N* is an integer, the function returns the value of :math:`\gamma`.
    The coterminal angle :math:`\gamma` is within the range :math:`[-\pi , \pi]`.

    Parameters
    ----------
    theta : array_like
        Input angles

    Returns
    -------
    numpy.ndarray
        Coterminal angles

    """
    coterminal = (theta + np.pi) % (2 * np.pi) - np.pi

    return coterminal


def define_plane_from_points(xyz1, xyz2, xyz3):
    """Compute constants defining a plane from a set of points.

    The equation defining a plane has the form :math:`ax+by+cz+d=0`.
    This utility returns the constants a, b, c and d defining the plane.

    Parameters
    ----------
    xyz1 : (3) numpy.ndarray
        First point needed to define the plane (x1, y1, z1)
    xyz2 : (3) numpy.ndarray
        Second point needed to define the plane (x2, y2, z2)
    xyz3 : (3) numpy.ndarray
        Third point needed to define the plane (x3, y3, z3)

    Returns
    -------
    a : float
    b : float
    c : float
    d : float

    """
    v12 = (xyz2 - xyz1) / np.sqrt(np.sum((xyz2 - xyz1) ** 2, axis=0))
    v13 = (xyz3 - xyz1) / np.sqrt(np.sum((xyz3 - xyz1) ** 2, axis=0))

    a, b, c = np.cross(v12, v13)
    d = -(a * xyz1[0] + b * xyz1[1] + c * xyz1[2])

    return a, b, c, d


################################################
#             DEPRECATED FUNCTIONS
################################################


diagEst = deprecate_function(estimate_diagonal, "diagEst", removal_version="0.19.0")
uniqueRows = deprecate_function(unique_rows, "uniqueRows", removal_version="0.19.0")
sdInv = deprecate_function(sdinv, "sdInv", removal_version="0.19.0", future_warn=True)
getSubArray = deprecate_function(
    get_subarray, "getSubArray", removal_version="0.19.0", future_warn=True
)
inv3X3BlockDiagonal = deprecate_function(
    inverse_3x3_block_diagonal,
    "inv3X3BlockDiagonal",
    removal_version="0.19.0",
    future_warn=True,
)
inv2X2BlockDiagonal = deprecate_function(
    inverse_2x2_block_diagonal,
    "inv2X2BlockDiagonal",
    removal_version="0.19.0",
    future_warn=True,
)
makePropertyTensor = deprecate_function(
    make_property_tensor,
    "makePropertyTensor",
    removal_version="0.19.0",
    future_warn=True,
)
invPropertyTensor = deprecate_function(
    inverse_property_tensor,
    "invPropertyTensor",
    removal_version="0.19.0",
    future_warn=True,
)



import scipy.sparse as sp

from .....electromagnetics.static.induced_polarization.simulation import (
    Simulation2DNodal as Sim,
)
from .....data import Data
import numpy as np
import numcodecs

numcodecs.blosc.use_threads = False

Sim.sensitivity_path = './sensitivity/'

from .simulation import (
    dask_getJtJdiag,
    dask_Jvec,
    dask_Jtvec,
    dask_dpred
)

from ..resistivity.simulation_2d import compute_J, dask_getSourceTerm

Sim.compute_J = compute_J
Sim.getSourceTerm = dask_getSourceTerm
Sim.getJtJdiag = dask_getJtJdiag
Sim.Jvec = dask_Jvec
Sim.Jtvec = dask_Jtvec
Sim.dpred = dask_dpred


def dask_fields(self, m=None, return_Ainv=False):
    if m is not None:
        self.model = m

    kys = self._quad_points
    f = self.fieldsPair(self)
    f._quad_weights = self._quad_weights

    Ainv_out = {}
    for iky, ky in enumerate(kys):
        A = self.getA(ky)
        Ainv = self.solver(A, **self.solver_opts)
        Ainv_out[iky] = self.solver(sp.csr_matrix(A.T), **self.solver_opts)
        RHS = self.getRHS(ky)
        f[:, self._solutionType, iky] = Ainv * RHS

        Ainv.clean()

    if self._scale is None:
        scale = Data(self.survey, np.ones(self.survey.nD))
        f_fwd = self.fields_to_space(f)
        # loop through receievers to check if they need to set the _dc_voltage
        for src in self.survey.source_list:
            for rx in src.receiver_list:
                if (
                        rx.data_type == "apparent_chargeability"
                        or self._data_type == "apparent_chargeability"
                ):
                    scale[src, rx] = 1.0 / rx.eval(src, self.mesh, f_fwd)
        self._scale = scale.dobs

    if return_Ainv:
        return f, Ainv_out
    else:
        return f, None


Sim.fields = dask_fields





import numpy as np
import scipy.sparse as sp

from ....utils import (
    mkvc,
    Zero,
    validate_type,
    validate_string,
    validate_active_indices,
)
from ....data import Data
from ....base import BaseElectricalPDESimulation
from .survey import Survey
from .fields import Fields3DCellCentered, Fields3DNodal
from .utils import _mini_pole_pole
from discretize.utils import make_boundary_bool


class BaseDCSimulation(BaseElectricalPDESimulation):
    """
    Base DC Problem
    """

    _mini_survey = None

    Ainv = None

    def __init__(
        self,
        mesh,
        survey=None,
        storeJ=False,
        miniaturize=False,
        surface_faces=None,
        **kwargs,
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.storeJ = storeJ
        self.surface_faces = surface_faces
        # Do stuff to simplify the forward and JTvec operation if number of dipole
        # sources is greater than the number of unique pole sources
        miniaturize = validate_type("miniaturize", miniaturize, bool)
        if miniaturize:
            self._dipoles, self._invs, self._mini_survey = _mini_pole_pole(self.survey)

    @property
    def survey(self):
        """The DC survey object.

        Returns
        -------
        SimPEG.electromagnetics.static.resistivity.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def storeJ(self):
        """Whether to store the sensitivity matrix

        Returns
        -------
        bool
        """
        return self._storeJ

    @storeJ.setter
    def storeJ(self, value):
        self._storeJ = validate_type("storeJ", value, bool)

    @property
    def surface_faces(self):
        """Array defining which boundary faces to interpret as surfaces of Neumann boundary

        DC problems will always enforce a Neumann boundary on surface interfaces.
        The default (available on semi-structured grids) assumes the top interface
        is the surface.

        Returns
        -------
        None or (n_bf, ) numpy.ndarray of bool
        """
        return self._surface_faces

    @surface_faces.setter
    def surface_faces(self, value):
        if value is not None:
            n_bf = self.mesh.boundary_faces.shape[0]
            value = validate_active_indices("surface_faces", value, n_bf)
        self._surface_faces = value

    def fields(self, m=None, calcJ=True):
        if m is not None:
            self.model = m

        f = self.fieldsPair(self)
        if self.Ainv is not None:
            self.Ainv.clean()
        A = self.getA()
        self.Ainv = self.solver(A, **self.solver_opts)
        RHS = self.getRHS()

        f[:, self._solutionType] = self.Ainv * RHS

        return f

    def getJ(self, m, f=None):
        if getattr(self, "_Jmatrix", None) is None:
            if f is None:
                f = self.fields(m)
            self._Jmatrix = self._Jtvec(m, v=None, f=f).T
        return self._Jmatrix

    def dpred(self, m=None, f=None):
        if self._mini_survey is not None:
            # Temporarily set self.survey to self._mini_survey
            survey = self.survey
            self.survey = self._mini_survey

        data = super().dpred(m=m, f=f)

        if self._mini_survey is not None:
            # reset survey
            self.survey = survey

        return self._mini_survey_data(data)

    def getJtJdiag(self, m, W=None, f=None):
        """
        Return the diagonal of JtJ
        """
        if getattr(self, "_jtjdiag", None) is None:
            J = self.getJ(m, f=f)

            if W is None:
                W = np.ones(J.shape[0])
            else:
                W = W.diagonal() ** 2

            diag = np.zeros(J.shape[1])
            for i in range(J.shape[0]):
                diag += (W[i]) * (J[i] * J[i])

            self._jtjdiag = diag
        return self._jtjdiag

    def Jvec(self, m, v, f=None):
        """
        Compute sensitivity matrix (J) and vector (v) product.
        """
        if f is None:
            f = self.fields(m)

        self.model = m

        if self.storeJ:
            J = self.getJ(m, f=f)
            return J.dot(v)

        self.model = m

        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        Jv = []
        for source in survey.source_list:
            u_source = f[source, self._solutionType]  # solution vector
            dA_dm_v = self.getADeriv(u_source, v)
            dRHS_dm_v = self.getRHSDeriv(source, v)
            du_dm_v = self.Ainv * (-dA_dm_v + dRHS_dm_v)
            for rx in source.receiver_list:
                df_dmFun = getattr(f, "_{0!s}Deriv".format(rx.projField), None)
                df_dm_v = df_dmFun(source, du_dm_v, v, adjoint=False)
                Jv.append(rx.evalDeriv(source, self.mesh, f, df_dm_v))
        Jv = np.hstack(Jv)
        return self._mini_survey_data(Jv)

    def Jtvec(self, m, v, f=None):
        """
        Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        """

        if f is None:
            f = self.fields(m)

        self.model = m

        if self.storeJ:
            J = self.getJ(m, f=f)
            return np.asarray(J.T.dot(v))

        return self._Jtvec(m, v=v, f=f)

    def _Jtvec(self, m, v=None, f=None):
        """
        Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        Full J matrix can be computed by inputing v=None
        """

        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        if v is not None:
            if isinstance(v, Data):
                v = v.dobs
            v = self._mini_survey_dataT(v)
            v = Data(survey, v)
            Jtv = np.zeros(m.size)
        else:
            # This is for forming full sensitivity matrix
            Jtv = np.zeros((self.model.size, survey.nD), order="F")
            istrt = int(0)
            iend = int(0)

        for source in survey.source_list:
            u_source = f[source, self._solutionType].copy()
            for rx in source.receiver_list:
                # wrt f, need possibility wrt m
                if v is not None:
                    PTv = rx.evalDeriv(
                        source, self.mesh, f, v[source, rx], adjoint=True
                    )
                else:
                    PTv = rx.evalDeriv(source, self.mesh, f).toarray().T

                df_duTFun = getattr(f, "_{0!s}Deriv".format(rx.projField), None)
                df_duT, df_dmT = df_duTFun(source, None, PTv, adjoint=True)

                ATinvdf_duT = self.Ainv * df_duT

                dA_dmT = self.getADeriv(u_source, ATinvdf_duT, adjoint=True)
                dRHS_dmT = self.getRHSDeriv(source, ATinvdf_duT, adjoint=True)
                du_dmT = -dA_dmT + dRHS_dmT
                if v is not None:
                    Jtv += (df_dmT + du_dmT).astype(float)
                else:
                    iend = istrt + rx.nD
                    if rx.nD == 1:
                        Jtv[:, istrt] = df_dmT + du_dmT
                    else:
                        Jtv[:, istrt:iend] = df_dmT + du_dmT
                    istrt += rx.nD

        if v is not None:
            return mkvc(Jtv)
        else:
            return (self._mini_survey_data(Jtv.T)).T

    def getSourceTerm(self):
        """
        Evaluates the sources, and puts them in matrix form
        :rtype: tuple
        :return: q (nC or nN, nSrc)
        """

        if getattr(self, "_q", None) is None:
            if self._mini_survey is not None:
                Srcs = self._mini_survey.source_list
            else:
                Srcs = self.survey.source_list

            if self._formulation == "EB":
                n = self.mesh.nN

            elif self._formulation == "HJ":
                n = self.mesh.nC

            q = np.zeros((n, len(Srcs)), order="F")

            for i, source in enumerate(Srcs):
                q[:, i] = source.eval(self)
            self._q = q
        return self._q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super().deleteTheseOnModelUpdate
        return toDelete + ["_Jmatrix", "_jtjdiag"]

    def _mini_survey_data(self, d_mini):
        if self._mini_survey is not None:
            out = d_mini[self._invs[0]]  # AM
            out[self._dipoles[0]] -= d_mini[self._invs[1]]  # AN
            out[self._dipoles[1]] -= d_mini[self._invs[2]]  # BM
            out[self._dipoles[0] & self._dipoles[1]] += d_mini[self._invs[3]]  # BN
        else:
            out = d_mini
        return out

    def _mini_survey_dataT(self, v):
        if self._mini_survey is not None:
            out = np.zeros(self._mini_survey.nD)
            # Need to use ufunc.at because there could be repeated indices
            # That need to be properly handled.
            np.add.at(out, self._invs[0], v)  # AM
            np.subtract.at(out, self._invs[1], v[self._dipoles[0]])  # AN
            np.subtract.at(out, self._invs[2], v[self._dipoles[1]])  # BM
            np.add.at(out, self._invs[3], v[self._dipoles[0] & self._dipoles[1]])  # BN
            return out
        else:
            out = v
        return out


class Simulation3DCellCentered(BaseDCSimulation):
    """
    3D cell centered DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "HJ"  # CC potentials means J is on faces
    fieldsPair = Fields3DCellCentered

    def __init__(self, mesh, survey=None, bc_type="Robin", **kwargs):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.bc_type = bc_type
        self.setBC()

    @property
    def bc_type(self):
        """Type of boundary condition to use for simulation.

        Returns
        -------
        {"Dirichlet", "Neumann", "Robin", "Mixed"}

        Notes
        -----
        Robin and Mixed are equivalent.
        """
        return self._bc_type

    @bc_type.setter
    def bc_type(self, value):
        self._bc_type = validate_string(
            "bc_type", value, ["Dirichlet", "Neumann", ("Robin", "Mixed")]
        )

    def getA(self, resistivity=None):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """

        D = self.Div
        G = self.Grad
        if resistivity is None:
            MfRhoI = self.MfRhoI
        else:
            MfRhoI = self.mesh.get_face_inner_product(resistivity, invert_matrix=True)
        A = D @ MfRhoI @ G

        if self.bc_type == "Neumann":
            if self.verbose:
                print("Perturbing first row of A to remove nullspace for Neumann BC.")

            # Handling Null space of A
            I, J, V = sp.find(A[0, :])
            for jj in J:
                A[0, jj] = 0.0
            A[0, 0] = 1.0

        return A

    def getADeriv(self, u, v, adjoint=False):
        if self.rhoMap is not None:
            D = self.Div
            G = self.Grad
            MfRhoIDeriv = self.MfRhoIDeriv

            if adjoint:
                return MfRhoIDeriv(G @ u, D.T @ v, adjoint)

            return D * (MfRhoIDeriv(G @ u, v, adjoint))
        return Zero()

    def getRHS(self):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm()

        return RHS

    def getRHSDeriv(self, source, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = source.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self):
        mesh = self.mesh
        V = sp.diags(mesh.cell_volumes)
        self.Div = V @ mesh.face_divergence
        self.Grad = self.Div.T

        if self.bc_type == "Dirichlet":
            if self.verbose:
                print(
                    "Homogeneous Dirichlet is the natural BC for this CC discretization."
                )
            # do nothing
            return
        elif self.bc_type == "Neumann":
            alpha, beta, gamma = 0, 1, 0
        else:
            boundary_faces = mesh.boundary_faces
            boundary_normals = mesh.boundary_face_outward_normals
            n_bf = len(boundary_faces)

            # Top gets 0 Nuemann
            alpha = np.zeros(n_bf)
            beta = np.ones(n_bf)
            gamma = 0

            # assume a source point at the middle of the top of the mesh
            middle = np.median(mesh.nodes, axis=0)
            top_v = np.max(mesh.nodes[:, -1])
            source_point = np.r_[middle[:-1], top_v]

            # Others: Robin: alpha * phi + d phi dn = 0
            # where alpha = 1 / r  * r_hat_dot_n
            # TODO: Implement Zhang et al. (1995)

            r_vec = boundary_faces - source_point
            r = np.linalg.norm(r_vec, axis=-1)
            r_hat = r_vec / r[:, None]
            r_dot_n = np.einsum("ij,ij->i", r_hat, boundary_normals)

            if self.surface_faces is None:
                # determine faces that are on the sides and bottom of the mesh...
                if mesh._meshType.lower() == "tree":
                    not_top = boundary_faces[:, -1] != top_v
                elif mesh._meshType.lower() in ["tensor", "curv"]:
                    # mesh faces are ordered, faces_x, faces_y, faces_z so...
                    if mesh.dim == 2:
                        is_b = make_boundary_bool(mesh.shape_faces_y)
                        is_t = np.zeros(mesh.shape_faces_y, dtype=bool, order="F")
                        is_t[:, -1] = True
                    else:
                        is_b = make_boundary_bool(mesh.shape_faces_z)
                        is_t = np.zeros(mesh.shape_faces_z, dtype=bool, order="F")
                        is_t[:, :, -1] = True
                    is_t = is_t.reshape(-1, order="F")[is_b]
                    not_top = np.ones(boundary_faces.shape[0], dtype=bool)
                    not_top[-len(is_t) :] = ~is_t
                    self.surface_faces = ~not_top
                else:
                    raise NotImplementedError(
                        f"Unable to infer surface boundaries for {type(mesh)}, please "
                        f"set the `surface_faces` property."
                    )
            else:
                not_top = ~self.surface_faces
            alpha[not_top] = (r_dot_n / r)[not_top]

        B, bc = mesh.cell_gradient_weak_form_robin(alpha, beta, gamma)
        # bc should always be 0 because gamma was always 0 above
        self.Grad = self.Grad - B


class Simulation3DNodal(BaseDCSimulation):
    """
    3D nodal DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "EB"  # N potentials means B is on faces
    fieldsPair = Fields3DNodal

    def __init__(self, mesh, survey=None, bc_type="Robin", **kwargs):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        # Not sure why I need to do this
        # To evaluate mesh.aveE2CC, this is required....
        if mesh._meshType == "TREE":
            mesh.nodal_gradient
        elif mesh._meshType == "CYL":
            bc_type = "Neumann"
        self.bc_type = bc_type
        self.setBC()

    @property
    def bc_type(self):
        """Type of boundary condition to use for simulation.

        Returns
        -------
        {"Neumann", "Robin", "Mixed"}

        Notes
        -----
        Robin and Mixed are equivalent.
        """
        return self._bc_type

    @bc_type.setter
    def bc_type(self, value):
        self._bc_type = validate_string(
            "bc_type", value, ["Neumann", ("Robin", "Mixed")]
        )

    def getA(self, resistivity=None):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = G.T MeSigma G
        """
        if resistivity is None:
            MeSigma = self.MeSigma
        else:
            MeSigma = self.mesh.get_edge_inner_product(1.0 / resistivity)
        Grad = self.mesh.nodal_gradient
        A = Grad.T.tocsr() @ MeSigma @ Grad

        if self.bc_type == "Neumann":
            # Handling Null space of A
            I, J, V = sp.find(A[0, :])
            for jj in J:
                A[0, jj] = 0.0
            A[0, 0] = 1.0
        else:
            # Dirichlet BC type should already have failed
            # Also, this will fail if sigma is anisotropic
            try:
                A = A + sp.diags(self._AvgBC @ self.sigma, format="csr")
            except ValueError as err:
                if len(self.sigma) != len(self.mesh):
                    raise NotImplementedError(
                        "Anisotropic conductivity is not supported for Robin boundary "
                        "conditions, please use 'Neumann'."
                    )
                else:
                    raise err

        return A

    def getADeriv(self, u, v, adjoint=False):
        """
        Product of the derivative of our system matrix with respect to the
        model and a vector
        """
        Grad = self.mesh.nodal_gradient
        if not adjoint:
            out = Grad.T @ self.MeSigmaDeriv(Grad @ u, v, adjoint)
        else:
            out = self.MeSigmaDeriv(Grad @ u, Grad @ v, adjoint)
        if self.bc_type != "Neumann" and self.sigmaMap is not None:
            if getattr(self, "_MBC_sigma", None) is None:
                self._MBC_sigma = self._AvgBC @ self.sigmaDeriv
            if not isinstance(u, Zero):
                u = u.flatten()
                if v.ndim > 1:
                    u = u[:, None]
                if not adjoint:
                    out += u * (self._MBC_sigma @ v)
                else:
                    out += self._MBC_sigma.T @ (u * v)
        return out

    def setBC(self):
        if self.bc_type == "Dirichlet":
            # do nothing
            raise ValueError(
                "Dirichlet conditions are not supported in the Nodal formulation"
            )
        elif self.bc_type == "Neumann":
            if self.verbose:
                print(
                    "Homogeneous Neumann is the natural BC for this nodal discretization."
                )
            return
        else:
            mesh = self.mesh
            # calculate alpha, beta, gamma at the boundary faces
            boundary_faces = mesh.boundary_faces
            boundary_normals = mesh.boundary_face_outward_normals
            n_bf = len(boundary_faces)

            # Top gets 0 Nuemann
            alpha = np.zeros(n_bf)
            # beta = np.ones(n_bf) = 1.0

            # not top get Robin condition
            # assume a source point at the middle of the top of the mesh
            middle = np.median(mesh.nodes, axis=0)
            top_v = np.max(mesh.nodes[:, -1])
            source_point = np.r_[middle[:-1], top_v]

            # Others: Robin: alpha * phi + d phi dn = 0
            # where alpha = 1 / r  * r_hat_dot_n
            # TODO: Implement Zhang et al. (1995)

            r_vec = boundary_faces - source_point
            r = np.linalg.norm(r_vec, axis=-1)
            r_hat = r_vec / r[:, None]
            r_dot_n = np.einsum("ij,ij->i", r_hat, boundary_normals)

            # determine faces that are on the sides and bottom of the mesh...
            if self.surface_faces is None:
                if mesh._meshType.lower() == "tree":
                    not_top = boundary_faces[:, -1] != top_v
                elif mesh._meshType.lower() in ["tensor", "curv"]:
                    # mesh faces are ordered, faces_x, faces_y, faces_z so...
                    if mesh.dim == 2:
                        is_b = make_boundary_bool(mesh.shape_faces_y)
                        is_t = np.zeros(mesh.shape_faces_y, dtype=bool, order="F")
                        is_t[:, -1] = True
                    else:
                        is_b = make_boundary_bool(mesh.shape_faces_z)
                        is_t = np.zeros(mesh.shape_faces_z, dtype=bool, order="F")
                        is_t[:, :, -1] = True
                    is_t = is_t.reshape(-1, order="F")[is_b]
                    not_top = np.ones(boundary_faces.shape[0], dtype=bool)
                    not_top[-len(is_t) :] = ~is_t
                else:
                    raise NotImplementedError(
                        f"Unable to infer surface boundaries for {type(mesh)}, please "
                        f"set the `surface_faces` property."
                    )
            else:
                not_top = ~self.surface_faces

            alpha[not_top] = (r_dot_n / r)[not_top]

            P_bf = self.mesh.project_face_to_boundary_face

            AvgN2Fb = P_bf @ self.mesh.average_node_to_face
            AvgCC2Fb = P_bf @ self.mesh.average_cell_to_face

            AvgCC2Fb = sp.diags(alpha * (P_bf @ self.mesh.face_areas)) @ AvgCC2Fb
            self._AvgBC = AvgN2Fb.T @ AvgCC2Fb

    def getRHS(self):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm()
        return RHS

    def getRHSDeriv(self, source, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = source.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()

    @property
    def _clear_on_sigma_update(self):
        """
        These matrices are deleted if there is an update to the conductivity
        model
        """
        return super()._clear_on_sigma_update + ["_MBC_sigma"]


Simulation3DCellCentred = Simulation3DCellCentered  # UK and US!



import numpy as np

from ....utils import sdiag, validate_string, validate_ndarray_with_shape

from ....survey import BaseTimeRx


class BaseRx(BaseTimeRx):
    """
    Base spectral IP receiver class

    Parameters
    ----------
    locations : (n_loc, dim) numpy.ndarray
        Receiver locations.
    times : numpy.array_like
        Time channels
    orientation : {``None``, 'x', 'y', 'z'}
        Receiver orientation
    data_type : {"volt", "apparent_chargeability"}
        Data type.
    projField : {"phi", "e", "j"}
        Fields solved on the mesh.
    """

    def __init__(
        self,
        locations=None,
        times=None,
        data_type="volt",
        orientation=None,
        projField="phi",
        **kwargs,
    ):
        super(BaseRx, self).__init__(locations=locations, times=times, **kwargs)

        self.orientation = orientation
        self.data_type = data_type
        self.projField = projField

    @property
    def orientation(self):
        """Orientation of the receiver.

        Returns
        -------
        str
            Orientation of the receiver. One of {None, 'x', 'y', 'z'}
        """
        return self._orientation

    @orientation.setter
    def orientation(self, var):
        if var is not None:
            var = validate_string("orientation", var, ("x", "y", "z")).lower()
        self._orientation = var

    @property
    def projField(self):
        """Fields on the mesh

        Returns
        -------
        str
            Fields defined on mesh. One of {"phi", "e", "j"}
        """
        return self._projField

    @projField.setter
    def projField(self, var):
        var = validate_string("projField", var, ("phi", "e", "j")).lower()
        self._projField = var

    @property
    def data_type(self):
        """Data type; i.e. "volt", "apparent_resistivity", "apparent_chargeability"

        Returns
        -------
        str
            Data type; i.e. "volt", "apparent_resistivity", "apparent_chargeability"
        """
        return self._data_type

    @data_type.setter
    def data_type(self, var):
        self._data_type = validate_string(
            "data_type",
            var,
            (
                ("volt", "potential", "potentials", "v", "voltage", "voltages"),
                (
                    "apparent_chargeability",
                    "apparent chargeability",
                    "appchargeability",
                    "apparentchargeability",
                    "apparent-chargeability",
                ),
            ),
        )

    # @property
    # def projField(self):
    #     """Field Type projection (e.g. e b ...)"""
    #     return self.knownRxTypes[self.rxType][0]

    @property
    def dc_voltage(self):
        """DC voltage

        Returns
        -------
        numpy.ndarray
            DC data for each receiver
        """
        return self._dc_voltage

    # def projected_grid(self, f):
    #     """Grid Location projection (e.g. Ex Fy ...)"""
    #     if self.orientation is not None:
    #         return f._GLoc(self.projField) + self.orientation
    #     return f._GLoc(self.projField)

    def getTimeP(self, times_all):
        """Returns the time projection matrix.

        This is not stored in memory, but is created on demand.

        Parameters
        ----------
        times_all : numpy.ndarray
            All times using the time-stepping

        Returns
        -------
        numpy.ndarray
            Time projection from time-steps to time channels
        """
        time_inds = np.in1d(times_all, self.times)
        return time_inds

    def eval(self, src, mesh, f):  # noqa: A003
        """Project fields to receivers to get data.

        Parameters
        ----------
        src : SimPEG.electromagnetics.static.spectral_induced_polarization.sources.BaseRx
            A spectral IP receiver
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        f : SimPEG.electromagnetic.static.spectral_induced_polarization.Fields
            The solution for the fields defined on the mesh

        Returns
        -------
        numpy.ndarray
            Fields projected to the receiver(s)
        """
        if self.orientation is not None:
            projected_grid = f._GLoc(self.projField) + self.orientation
        else:
            projected_grid = f._GLoc(self.projField)

        P = self.getP(mesh, projected_grid)
        proj_f = self.projField
        if proj_f == "phi":
            proj_f = "phiSolution"
        return P * f[src, proj_f]

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        """Derivative of projected fields with respect to the inversion model times a vector.

        Parameters
        ----------
        src : SimPEG.electromagnetics.static.spectral_induced_polarization.sources.BaseRx
            A spectral IP receiver
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        f : SimPEG.electromagnetic.static.spectral_induced_polarization.Fields
            The solution for the fields defined on the mesh
        v : numpy.ndarray
            A vector
        adjoint : bool, default = ``False``
            If ``True``, return the adjoint

        Returns
        -------
        numpy.ndarray
            derivative of fields times a vector projected to the receiver(s)
        """
        if self.orientation is not None:
            projected_grid = f._GLoc(self.projField) + self.orientation
        else:
            projected_grid = f._GLoc(self.projField)

        P = self.getP(mesh, projected_grid)
        if not adjoint:
            return P * v
        elif adjoint:
            return P.T * v


class Dipole(BaseRx):
    """
    Spectral IP dipole receiver class

    Parameters
    ----------
    locations_m : (n_loc, dim) numpy.ndarray
        M electrode locations; remember to set 'locations_n' keyword argument to define N electrode locations.
    locations_n : (n_loc, dim) numpy.ndarray
        N electrode locations; remember to set 'locations_m' keyword argument to define M electrode locations.
    locations : list or tuple of length 2 of numpy.ndarray
        M and N electrode locations. In this case, do not set the 'locations_m' and 'locations_n'
        keyword arguments. And we supply a list or tuple of the form [locations_m, locations_n].
    orientation : {``None``, 'x', 'y', 'z'}
        Receiver orientation.
    data_type : {"volt", "apparent_chargeability"}
        Data type.
    """

    def __init__(
        self, locations_m=None, locations_n=None, times=None, locations=None, **kwargs
    ):
        # if locations_m set, then use locations_m, locations_n
        if locations_m is not None:
            if locations_n is None:
                raise ValueError(
                    "For a dipole source both locations_m and locations_n "
                    "must be set"
                )

            if locations is not None:
                raise ValueError(
                    "Cannot set both locations and locations_m, locations_n. "
                    "Please provide either locations=(locations_m, locations_n) "
                    "or both locations_m=locations_m, locations_n=locations_n"
                )

            locations = [np.atleast_2d(locations_m), np.atleast_2d(locations_n)]

        if locations is None:
            raise AttributeError(
                "Receiver cannot be instantiated without assigning 'locations'."
                "Please provide either locations=(locations_m, locations_n) "
                "or both locations_m=locations_m, locations_n=locations_n"
            )

        super().__init__(locations=locations, times=times, **kwargs)

    # @property
    # def locations_m(self):
    #     """Locations of the M-electrodes"""
    #     return self.locations[0]

    # @property
    # def locations_n(self):
    #     """Locations of the N-electrodes"""
    #     return self.locations[1]

    @property
    def locations(self):
        """M and N electrode locations

        Returns
        -------
        list of 2 (n_loc, dim) np.ndarray
            M and N electrode locations.
        """
        return self._locations

    @locations.setter
    def locations(self, locs):
        if len(locs) != 2:
            raise ValueError(
                "locations must be a list or tuple of length 2: "
                "[locations_m, locations_n]. The input locations has "
                f"length {len(locs)}"
            )

        locs_m = validate_ndarray_with_shape("locations", locs[0], ("*", "*"))
        locs_n = validate_ndarray_with_shape("locations", locs[1], ("*", "*"))

        locs = [locs_m, locs_n]

        # check the size of locations_m, locations_n
        if locs[0].shape != locs[1].shape:
            raise ValueError(
                f"locations_m (shape: {locs[0].shape}) and "
                f"locations_n (shape: {locs[1].shape}) need to be "
                f"the same size"
            )

        self._locations = locs

    @property
    def locations_m(self):
        """Locations of the M-electrodes

        Returns
        -------
        (n, dim) numpy.ndarray
            Locations of the M-electrodes
        """
        return self.locations[0]

    @property
    def locations_n(self):
        """Locations of the N-electrodes

        Returns
        -------
        (n, dim) numpy.ndarray
            Locations of the N-electrodes
        """
        return self.locations[1]

    # this should probably be updated to n_receivers...
    @property
    def nD(self):
        """Number of data associate with the receiver(s).

        Returns
        -------
        int
            Number of data associated with the receiver(s).
        """
        return self.locations[0].shape[0]

    def getP(self, mesh, projected_grid):
        """
        Get projection matrix from mesh to receivers

        Parameters
        ----------
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        projected_grid : str
            Tensor locations on the mesh being interpolated from. *projected_grid* must be one of:

            - 'Ex', 'edges_x'           -> x-component of field defined on x edges
            - 'Ey', 'edges_y'           -> y-component of field defined on y edges
            - 'Ez', 'edges_z'           -> z-component of field defined on z edges
            - 'Fx', 'faces_x'           -> x-component of field defined on x faces
            - 'Fy', 'faces_y'           -> y-component of field defined on y faces
            - 'Fz', 'faces_z'           -> z-component of field defined on z faces
            - 'N', 'nodes'              -> scalar field defined on nodes
            - 'CC', 'cell_centers'      -> scalar field defined on cell centers
            - 'CCVx', 'cell_centers_x'  -> x-component of vector field defined on cell centers
            - 'CCVy', 'cell_centers_y'  -> y-component of vector field defined on cell centers
            - 'CCVz', 'cell_centers_z'  -> z-component of vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            P, the interpolation matrix

        """
        if mesh in self._Ps:
            return self._Ps[mesh]

        P0 = mesh.get_interpolation_matrix(self.locations[0], projected_grid)
        P1 = mesh.get_interpolation_matrix(self.locations[1], projected_grid)
        P = P0 - P1

        if self.data_type == "apparent_resistivity":
            P = sdiag(1.0 / self.geometric_factor) * P
        elif self.data_type == "apparent_chargeability":
            P = sdiag(1.0 / self.dc_voltage) * P

        if self.storeProjections:
            self._Ps[mesh] = P

        return P


class Pole(BaseRx):
    """
    Spectral IP pole receiver class

    Parameters
    ----------
    locations : (n_loc, dim) numpy.ndarray
        Receiver locations.
    orientation : {``None``, 'x', 'y', 'z'}
        Receiver orientation.
    data_type : {"volt", "apparent_resistivity", "apparent_chargeability"}
        Data type.
    """

    # this should probably be updated to n_receivers...
    @property
    def nD(self):
        """Number of data associate with the receiver(s).

        Returns
        -------
        int
            Number of data associated with the receiver(s).
        """
        return self.locations.shape[0]

    def getP(self, mesh, projected_grid):
        """
        Get projection matrix from mesh to receivers

        Parameters
        ----------
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        projected_grid : str
            Tensor locations on the mesh being interpolated from. *projected_grid* must be one of:

            - 'Ex', 'edges_x'           -> x-component of field defined on x edges
            - 'Ey', 'edges_y'           -> y-component of field defined on y edges
            - 'Ez', 'edges_z'           -> z-component of field defined on z edges
            - 'Fx', 'faces_x'           -> x-component of field defined on x faces
            - 'Fy', 'faces_y'           -> y-component of field defined on y faces
            - 'Fz', 'faces_z'           -> z-component of field defined on z faces
            - 'N', 'nodes'              -> scalar field defined on nodes
            - 'CC', 'cell_centers'      -> scalar field defined on cell centers
            - 'CCVx', 'cell_centers_x'  -> x-component of vector field defined on cell centers
            - 'CCVy', 'cell_centers_y'  -> y-component of vector field defined on cell centers
            - 'CCVz', 'cell_centers_z'  -> z-component of vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            P, the interpolation matrix
        """
        if mesh in self._Ps:
            return self._Ps[mesh]

        P = mesh.get_interpolation_matrix(self.locations, projected_grid)

        if self.data_type == "apparent_resistivity":
            P = sdiag(1.0 / self.geometric_factor) * P
        elif self.data_type == "apparent_chargeability":
            P = sdiag(1.0 / self.dc_voltage) * P

        if self.storeProjections:
            self._Ps[mesh] = P

        return P



import numpy as np
from .maps import IdentityMap


class Model(np.ndarray):
    def __new__(cls, input_array, mapping=None):
        assert isinstance(mapping, IdentityMap), "mapping must be a SimPEG.Mapping"
        assert isinstance(input_array, np.ndarray), "input_array must be a numpy array"
        assert len(input_array.shape) == 1, "input_array must be a 1D vector"
        obj = np.asarray(input_array).view(cls)
        obj._mapping = mapping
        if not obj.size == mapping.nP:
            raise Exception("Incorrect size for array.")
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._mapping = getattr(obj, "_mapping", None)

    @property
    def mapping(self):
        return self._mapping

    @property
    def transform(self):
        if getattr(self, "_transform", None) is None:
            self._transform = self.mapping * self.view(np.ndarray)
        return self._transform

    @property
    def transformDeriv(self):
        if getattr(self, "_transformDeriv", None) is None:
            self._transformDeriv = self.mapping.deriv(self.view(np.ndarray))
        return self._transformDeriv



import numpy as np
from scipy.constants import mu_0, pi, epsilon_0
from SimPEG import utils


def hzAnalyticDipoleF(r, freq, sigma, secondary=True, mu=mu_0):
    """
    The analytical expression is given in Equation 4.56 in Ward and Hohmann,
    1988, and the example reproduces their Figure 4.2.


    .. plot::

        import numpy as np
        import matplotlib.pyplot as plt
        from SimPEG import electromagnetics as EM
        freq = np.logspace(-1, 5, 301)
        test = EM.analytics.h[2]AnalyticDipoleF(
                100, freq, 0.01, secondary=False)
        plt.loglog(freq, test.real, 'C0-', label='Real')
        plt.loglog(freq, -test.real, 'C0--')
        plt.loglog(freq, test.imag, 'C1-', label='Imaginary')
        plt.loglog(freq, -test.imag, 'C1--')
        plt.title('Response at $r=100$ m')
        plt.xlim([1e-1, 1e5])
        plt.ylim([1e-12, 1e-6])
        plt.xlabel('Frequency (Hz)')
        plt.ylabel('$H_z$ (A/m)')
        plt.legend(loc=6)
        plt.show()


    **Reference**

    - Ward, S. H., and G. W. Hohmann, 1988, Electromagnetic theory for
      geophysical applications, Chapter 4 of Electromagnetic Methods in Applied
      Geophysics: SEG, Investigations in Geophysics No. 3, 130--311; DOI:
      `10.1190/1.9781560802631.ch4
      <https://doi.org/10.1190/1.9781560802631.ch4>`_.

    """
    r = np.abs(r)
    k = np.sqrt(-1j * 2.0 * np.pi * freq * mu * sigma)

    m = 1
    front = m / (2.0 * np.pi * (k**2) * (r**5))
    back = 9 - (
        9 + 9j * k * r - 4 * (k**2) * (r**2) - 1j * (k**3) * (r**3)
    ) * np.exp(-1j * k * r)
    hz = front * back

    if secondary:
        hp = -1 / (4 * np.pi * r**3)
        hz = hz - hp

    if hz.ndim == 1:
        hz = utils.mkvc(hz, 2)

    return hz


def MagneticDipoleWholeSpace(
    XYZ, srcLoc, sig, f, moment, fieldType="b", mu_r=1, eps_r=1, **kwargs
):
    """
    Analytical solution for a dipole in a whole-space.

    The analytical expression is given in Equation 2.57 in Ward and Hohmann,
    1988, and the example reproduces their Figure 2.2.

    TODOs:
        - set it up to instead take a mesh & survey
        - add divide by zero safety


    .. plot::

        import numpy as np
        from SimPEG import electromagnetics as EM
        import matplotlib.pyplot as plt
        from scipy.constants import mu_0
        freqs = np.logspace(-2, 5, 301)
        Bx, By, Bz = EM.analytics.FDEM.MagneticDipoleWholeSpace(
                [0, 100, 0], [0, 0, 0], 1e-2, freqs, moment='Z')
        plt.figure()
        plt.loglog(freqs, Bz.real/mu_0, 'C0', label='Real')
        plt.loglog(freqs, -Bz.real/mu_0, 'C0--')
        plt.loglog(freqs, Bz.imag/mu_0, 'C1', label='Imaginary')
        plt.loglog(freqs, -Bz.imag/mu_0, 'C1--')
        plt.legend()
        plt.xlim([1e-2, 1e5])
        plt.ylim([1e-13, 1e-6])
        plt.show()

    **Reference**

    - Ward, S. H., and G. W. Hohmann, 1988, Electromagnetic theory for
      geophysical applications, Chapter 4 of Electromagnetic Methods in Applied
      Geophysics: SEG, Investigations in Geophysics No. 3, 130--311; DOI:
      `10.1190/1.9781560802631.ch4
      <https://doi.org/10.1190/1.9781560802631.ch4>`_.

    """

    orient = kwargs.pop("orientation", None)
    if orient is not None:
        raise TypeError(
            "orientation kwarg has been removed, please use the moment argument",
        )
        magnitude = moment
        moment = orient
    else:
        magnitude = 1
    mu = kwargs.pop("mu", None)
    if mu is not None:
        raise TypeError("mu kwarg has been removed, please use the mu_r argument.")
        mu_r = mu / mu_0

    mu = mu_0 * mu_r
    eps = epsilon_0 * eps_r
    w = 2 * np.pi * f

    if isinstance(moment, str):
        if moment == "X":
            mx, my, mz = 1.0, 0.0, 0.0
        elif moment == "Y":
            mx, my, mz = 0.0, 1.0, 0.0
        elif moment == "Z":
            mx, my, mz = 0.0, 0.0, 1.0
        else:
            raise NotImplementedError("String type for moment not recognized")
        mx, my, mz = mx * magnitude, my * magnitude, mz * magnitude
    else:
        mx, my, mz = moment[0], moment[1], moment[2]

    XYZ = utils.as_array_n_by_dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    k = np.sqrt(-1j * w * mu * sig + w**2 * mu * eps)
    kr = k * r

    if fieldType in ["h", "b"]:
        front = 1 / (4.0 * pi * r**3.0) * np.exp(-1j * kr)
        mid = -(kr**2.0) + 3.0 * 1j * kr + 3.0

        Fx = front * (
            mx * ((dx / r) ** 2.0 * mid + (kr**2.0 - 1j * kr - 1.0))
            + my * ((dy * dx / r**2.0) * mid)
            + mz * ((dx * dz / r**2.0) * mid)
        )

        Fy = front * (
            mx * ((dx * dy / r**2.0) * mid)
            + my * ((dy / r) ** 2.0 * mid + (kr**2.0 - 1j * kr - 1.0))
            + mz * ((dy * dz / r**2.0) * mid)
        )

        Fz = front * (
            mx * ((dx * dz / r**2.0) * mid)
            + my * ((dy * dz / r**2.0) * mid)
            + mz * ((dz / r) ** 2.0 * mid + (kr**2.0 - 1j * kr - 1.0))
        )

        if fieldType == "b":
            Fx, Fy, Fz = mu * Fx, mu * Fy, mu * Fz

    elif fieldType == "e":
        front = 1j * w * mu * (1 + 1j * kr) / (4.0 * pi * r**3.0) * np.exp(-1j * kr)

        Fx = front * (my * (dz / r) + mz * (-dy / r))

        Fy = front * (mx * (-dz / r) + mz * (dx / r))

        Fz = front * (mx * (dy / r) + my * (-dx / r))

    return Fx, Fy, Fz


def ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, moment="X", fieldType="e", mu_r=1, eps_r=1, **kwargs
):
    orient = kwargs.pop("orientation", None)
    if orient is not None:
        raise TypeError(
            "orientation kwarg has been removed, please use the moment argument."
        )
    mu = kwargs.pop("mu", None)
    if mu is not None:
        raise TypeError("mu kwarg has been removed, please use the mu_r argument.")
    cur = kwargs.pop("current", None)
    if cur is not None:
        raise TypeError(
            "current kwarg has been removed, please use the moment argument.",
        )
    else:
        magnitude = 1
    length = kwargs.pop("length", None)
    if length is not None:
        raise TypeError(
            "length kwarg has been removed, please use the moment argument."
        )

    mu = mu_0 * mu_r
    eps = epsilon_0 * eps_r
    w = 2 * np.pi * f

    if isinstance(moment, str):
        if moment.upper() == "X":
            mx, my, mz = 1.0, 0.0, 0.0
        elif moment.upper() == "Y":
            mx, my, mz = 0.0, 1.0, 0.0
        elif moment.upper() == "Z":
            mx, my, mz = 0.0, 0.0, 1.0
        else:
            raise NotImplementedError("String type for moment not recognized")
        mx, my, mz = mx * magnitude, my * magnitude, mz * magnitude

    else:
        mx, my, mz = moment[0], moment[1], moment[2]

    XYZ = utils.as_array_n_by_dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    k = np.sqrt(-1j * w * mu * sig + w**2 * mu * eps)
    kr = k * r

    if fieldType == "e":
        front = 1 / (4.0 * np.pi * sig * r**3) * np.exp(-1j * k * r)
        mid = -(k**2) * r**2 + 3 * 1j * k * r + 3

        Fx = front * (
            mx * ((dx**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
            + my * (dy * dx / r**2) * mid
            + mz * (dz * dx / r**2) * mid
        )

        Fy = front * (
            mx * (dx * dy / r**2) * mid
            + my * ((dy**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
            + mz * (dz * dy / r**2) * mid
        )

        Fz = front * (
            mx * (dx * dz / r**2) * mid
            + my * (dy * dz / r**2) * mid
            + mz * ((dz**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
        )

    elif fieldType in ["h", "b"]:
        front = (1 + 1j * kr) / (4.0 * np.pi * r**2) * np.exp(-1j * k * r)

        Fx = front * (my * (dz / r) + mz * (-dy / r))

        Fy = front * (mx * (-dz / r) + mz * (dx / r))

        Fz = front * (mx * (dy / r) + my * (-dx / r))

        if fieldType == "b":
            Fx, Fy, Fz = mu * Fx, mu * Fy, mu * Fz

    return Fx, Fy, Fz



"""
=========================================================================
Magnetics Simulation (:mod:`SimPEG.potential_fields.magnetics`)
=========================================================================
.. currentmodule:: SimPEG.potential_fields.magnetics

About ``magnetics``

Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation3DIntegral
  SimulationEquivalentSourceLayer
  Simulation3DDifferential

Survey, Source and Receiver Classes
===================================

.. autosummary::
  :toctree: generated/

  Point
  UniformBackgroundField
  Survey

Analytics
=========
.. autosummary::
  :toctree: generated/

  analytics.IDTtoxyz
  analytics.MagSphereAnaFun
  analytics.MagSphereAnaFunA
  analytics.MagSphereFreeSpace
"""
from . import survey
from . import sources
from . import receivers
from . import analytics
from . import simulation

from .simulation import (
    Simulation3DIntegral,
    SimulationEquivalentSourceLayer,
    Simulation3DDifferential,
)
from .survey import Survey
from .sources import SourceField, UniformBackgroundField
from .receivers import Point



import numpy as np
from scipy.constants import epsilon_0
from scipy.constants import mu_0
from SimPEG.electromagnetics.utils import k, omega

__all__ = ["MT_LayeredEarth"]


# Evaluate Impedance Z of a layer
_ImpZ = lambda f, mu, k: omega(f) * mu / k

# Complex Cole-Cole Conductivity - EM utils
_PCC = lambda siginf, m, t, c, f: siginf * (
    1.0 - (m / (1.0 + (1j * omega(f) * t) ** c))
)

# matrix P relating Up and Down components with E and H fields
_P = lambda z: np.array(
    [
        [
            1.0,
            1,
        ],
        [-1.0 / z, 1.0 / z],
    ]
)
_Pinv = lambda z: np.array([[1.0, -z], [1.0, z]]) / 2.0

# matrix T for transition of Up and Down components accross a layer
_T = lambda h, k: np.array(
    [[np.exp(1j * k * h), 0.0], [0.0, np.exp(-1j * k * h)]],
)
_Tinv = lambda h, k: np.array(
    [[np.exp(-1j * k * h), 0.0], [0.0, np.exp(1j * k * h)]],
)


# Propagate Up and Down component for a certain frequency & evaluate E and H field
def _Propagate(f, thickness, sig, chg, taux, c, mu_r, eps_r, n):
    if isinstance(eps_r, float):
        epsmodel = np.ones_like(sig) * eps_r
    else:
        epsmodel = eps_r

    if isinstance(mu_r, float):
        mumodel = np.ones_like(sig) * mu_r
    else:
        epsmodel = mu_r

    sigcm = np.zeros_like(sig, dtype="complex_")
    if chg == 0.0 or taux == 0.0 or c == 0.0:
        sigcm = sig
    else:
        for j in range(1, len(sigcm)):
            sigcm[j] = _PCC(sig[j], chg[j], taux[j], c[j], f)

    sigcm = np.append(np.r_[0.0], sigcm)
    mu = np.append(np.r_[1.0], mumodel) * mu_0
    eps = np.append(np.r_[1.0], epsmodel) * epsilon_0
    H = np.append(np.r_[1.2 * (1e5)], thickness)

    K = k(f, sigcm, mu, eps)
    Z = _ImpZ(f, mu, K)

    EH = np.zeros((2, n + 1), dtype="complex_")
    UD = np.zeros((2, n + 1), dtype="complex_")

    UD[1, -1] = 1.0

    for i in range(-2, -(n + 2), -1):
        UD[:, i] = _Tinv(H[i + 1], K[i]) @ _Pinv(Z[i]) @ _P(Z[i + 1]) @ UD[:, i + 1]
        UD = UD / ((np.abs(UD[0, :] + UD[1, :])).max())

    for j in range(0, n + 1):
        EH[:, j] = (
            np.array(
                [
                    [
                        1.0,
                        1,
                    ],
                    [-1.0 / Z[j], 1.0 / Z[j]],
                ]
            )
            @ UD[:, j]
        )

    return UD, EH, Z, K


# Utils to compute the apparent impedance over a layered Earth Model
def MT_LayeredEarth(
    freq,
    thickness,
    sig,
    return_type="Res-Phase",
    chg=0.0,
    tau=0.0,
    c=0.0,
    mu_r=1.0,
    eps_r=1.0,
):
    """
    This code compute the analytic response of a n-layered Earth to a plane wave (Magnetotellurics).
    All physical properties arrays convention describes the layers parameters from the top layer to the bottom layer.
    The solution is first developed in Ward and Hohmann 1988.
    See also http://em.geosci.xyz/content/maxwell3_fdem/natural_sources/MT_N_layered_Earth.html

    :param freq: the frequency at which we take the measurements
    :type freq: float or numpy.ndarray
    :param thickness: thickness of the Earth layers in meters, size is len(sig)-1. The last one is already considered infinite. For 1-layer Earth, thickness = None or 0.
    :type thickness: float or numpy.ndarray
    :param sig: electric conductivity of the Earth layers in S/m
    :type sig: float or numpy.ndarray
    :param str return_type: Output return_type. 'Res-Phase' returns apparent resisitivity and Phase. 'Impedance' returns the complex Impedance
    :param numpy.ndarray chg: Cole-Cole Parameters for chargeable layers: chargeability
    :param numpy.ndarray tau: Cole-Cole Parameters for chargeable layers: time decay constant
    :param numpy.ndarray c: Cole-Cole Parameters for chargeable layers: geometric factor
    :param mu_r: relative magnetic permeability
    :type mu_r: float or numpy.ndarray
    :param eps_r: relative dielectric permittivity
    :type eps_r: float or numpy.ndarray
    """

    if isinstance(freq, float):
        F = np.r_[freq]
    else:
        F = freq

    if isinstance(sig, float):
        sigmodel = np.r_[sig]
    else:
        sigmodel = sig

    if isinstance(thickness, float):
        if thickness == 0.0:
            thickmodel = np.empty(0)
        else:
            thickmodel = np.r_[thickness]
    elif thickness is None:
        thickmodel = np.empty(0)
    else:
        thickmodel = thickness

    # Count the number of layers
    nlayer = len(sigmodel)

    Res = np.zeros_like(F)
    Phase = np.zeros_like(F)
    App_ImpZ = np.zeros_like(F, dtype="complex_")

    for i in range(0, len(F)):
        _, EH, _, _ = _Propagate(
            F[i], thickmodel, sigmodel, chg, tau, c, mu_r, eps_r, nlayer
        )

        App_ImpZ[i] = EH[0, 1] / EH[1, 1]

        Res[i] = np.abs(App_ImpZ[i]) ** 2.0 / (mu_0 * omega(F[i]))
        Phase[i] = np.angle(App_ImpZ[i], deg=True)

    if return_type == "Res-Phase":
        return Res, Phase

    elif return_type == "Impedance":
        return App_ImpZ


def _run():
    # nlayer=1
    F0 = 1.0
    H0 = None
    H01 = 0.0
    sign0 = 0.1

    # nlayer = 2
    F1 = np.r_[1e-5, 1e3]
    H1 = 200.0
    sign1 = np.r_[0.1, 1.0]

    # nlayer1 = 3
    F2 = 1e-3
    H2 = np.r_[200.0, 50.0]
    sign2 = np.r_[0.01, 1.0, 0.1]
    fm = "Impedance"

    Res, Phase = MT_LayeredEarth(F0, H0, sign0)
    print(Res, Phase)
    Res, Phase = MT_LayeredEarth(F0, H01, sign0)
    print(Res, Phase)
    Res, Phase = MT_LayeredEarth(F1, H1, sign1)
    print(Res, Phase)
    appimp = MT_LayeredEarth(F2, H2, sign2, return_type=fm)
    print(appimp)


if __name__ == "__main__":
    _run()



from ... import survey
from ...utils import validate_string, validate_type, validate_direction
import warnings
from discretize.utils import Zero


class BaseRx(survey.BaseRx):
    """Base FDEM receivers class.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'z', 'x', 'y'} or numpy.ndarray
        Receiver orientation.
    component : {'real', 'imag', 'both', 'complex'}
        Component of the receiver; i.e. 'real' or 'imag'. The options 'both' and
        'complex' are only available for the 1D layered simulations.
    data_type : {'field', 'ppm'}
        Data type observed by the receiver, either field, or ppm secondary
        of the total field.
    use_source_receiver_offset : bool, optional
        Whether to interpret the receiver locations as defining the source and receiver
        offset.
    """

    def __init__(
        self,
        locations,
        orientation="z",
        component="real",
        data_type="field",
        use_source_receiver_offset=False,
        **kwargs,
    ):
        proj = kwargs.pop("projComp", None)
        if proj is not None:
            warnings.warn(
                "'projComp' overrides the 'orientation' property which automatically"
                " handles the projection from the mesh the receivers!!! "
                "'projComp' is deprecated and will be removed in SimPEG 0.19.0."
            )
            self.projComp = proj

        self.orientation = orientation
        self.component = component
        self.data_type = data_type
        self.use_source_receiver_offset = use_source_receiver_offset

        super().__init__(locations, **kwargs)

    @property
    def orientation(self):
        """Orientation of the receiver.

        Returns
        -------
        numpy.ndarray
        """
        return self._orientation

    @orientation.setter
    def orientation(self, var):
        self._orientation = validate_direction("orientation", var, dim=3)

    @property
    def component(self):
        """Data component; i.e. real or imaginary.

        Returns
        -------
        str : {'real', 'imag', 'both', 'complex'}
            Component of the receiver; i.e. 'real' or 'imag'. The options 'both' and
            'complex' are only available for the 1D layered simulations.
        """
        return self._component

    @component.setter
    def component(self, val):
        self._component = validate_string(
            "component",
            val,
            (
                ("real", "re", "in-phase", "in phase"),
                (
                    "imag",
                    "imaginary",
                    "im",
                    "out-of-phase",
                    "out of phase",
                    "quadrature",
                ),
                "both",
                "complex",
            ),
        )

    @property
    def data_type(self):
        """The type of data for this receiver.

        The data type is either a field measurement or a part per million (ppm) measurement
        of the primary field.

        Returns
        -------
        str : {'field', 'ppm'}

        Notes
        -----
        This is currently only implemented for the 1D layered simulations.
        """
        return self._data_type

    @data_type.setter
    def data_type(self, val):
        self._data_type = validate_string(
            "data_type", val, string_list=("field", "ppm")
        )

    @property
    def use_source_receiver_offset(self):
        """Use source-receiver offset.

        Whether to interpret the location as a source-receiver offset.

        Returns
        -------
        bool

        Notes
        -----
        This is currently only implemented for the 1D layered code.
        """
        return self._use_source_receiver_offset

    @use_source_receiver_offset.setter
    def use_source_receiver_offset(self, val):
        self._use_source_receiver_offset = validate_type(
            "use_source_receiver_offset", val, bool
        )

    def getP(self, mesh, projected_grid):
        """Get projection matrix from mesh to receivers

        Parameters
        ----------
        mesh : discretize.BaseMesh
            A discretize mesh
        projected_grid : str
            Define what part of the mesh (i.e. edges, faces, centers, nodes) to
            project from. Must be one of::

                'E', 'edges_'           -> field defined on edges
                'F', 'faces_'           -> field defined on faces
                'CCV', 'cell_centers_'  -> vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            P, the interpolation matrix
        """
        if (mesh, projected_grid) in self._Ps:
            return self._Ps[(mesh, projected_grid)]

        P = Zero()
        for strength, comp in zip(self.orientation, ["x", "y", "z"]):
            if strength != 0.0:
                P = P + strength * mesh.get_interpolation_matrix(
                    self.locations, projected_grid + comp
                )

        if self.storeProjections:
            self._Ps[(mesh, projected_grid)] = P
        return P

    def eval(self, src, mesh, f):  # noqa: A003
        """Project fields from the mesh to the receiver(s).

        Parameters
        ----------
        src : SimPEG.electromagnetics.frequency_domain.sources.BaseFDEMSrc
            A frequency-domain EM source
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        f : SimPEG.electromagnetic.frequency_domain.fields.FieldsFDEM
            The solution for the fields defined on the mesh

        Returns
        -------
        numpy.ndarray
            Fields projected to the receiver(s)
        """
        projected_grid = f._GLoc(self.projField)
        P = self.getP(mesh, projected_grid)
        f_part_complex = f[src, self.projField]
        f_part = getattr(f_part_complex, self.component)  # real or imag component

        return P * f_part

    def evalDeriv(self, src, mesh, f, du_dm_v=None, v=None, adjoint=False):
        """Derivative of the projected fields with respect to the model, times a vector.

        Parameters
        ----------
        src : SimPEG.electromagnetics.frequency_domain.sources.BaseFDEMSrc
            A frequency-domain EM source
        mesh : discretize.base.BaseMesh
            The mesh on which the discrete set of equations is solved
        f : SimPEG.electromagnetic.frequency_domain.fields.FieldsFDEM
            The solution for the fields defined on the mesh
        du_dm_v : numpy.ndarray
            The derivative of the fields on the mesh with respect to the model,
            times a vector.
        v : numpy.ndarray, optional
            The vector which being multiplied
        adjoint : bool
            If ``True``, return the ajoint

        Returns
        -------
        numpy.ndarray
            The derivative times a vector at the receiver(s)
        """

        df_dmFun = getattr(f, "_{0}Deriv".format(self.projField), None)

        assert v is not None, "v must be provided to compute the deriv or adjoint"

        projected_grid = f._GLoc(self.projField)
        P = self.getP(mesh, projected_grid)

        if not adjoint:
            assert (
                du_dm_v is not None
            ), "du_dm_v must be provided to evaluate the receiver deriv"
            df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
            Pv_complex = P * df_dm_v
            Pv = getattr(Pv_complex, self.component)

            return Pv

        elif adjoint:
            PTv_real = P.T * v

            if self.component == "imag":
                PTv = -1j * PTv_real
            elif self.component == "real":
                PTv = PTv_real.astype(complex)
            else:
                raise NotImplementedError("must be real or imag")

            df_duT, df_dmT = df_dmFun(src, None, PTv, adjoint=True)
            # if self.component == "imag":  # conjugate
            #     df_duT *= -1
            #     df_dmT *= -1

            return df_duT, df_dmT

    @property
    def nD(self):
        if self.component == "both":
            return int(self.locations.shape[0] * 2)
        else:
            return self.locations.shape[0]


class PointElectricField(BaseRx):
    """Measure FDEM electric field at a point.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation.
    component : {'real', 'imag'}
        Real or imaginary component.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "e"
        super().__init__(locations, orientation, component, **kwargs)


class PointMagneticFluxDensity(BaseRx):
    """Measure FDEM total field magnetic flux density at a point.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation.
    component : {'real', 'imag'}
        Real or imaginary component.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "b"
        super().__init__(locations, orientation, component, **kwargs)


class PointMagneticFluxDensitySecondary(BaseRx):
    """Measure FDEM secondary magnetic flux density at a point.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation.
    component : {'real', 'imag'}
        Real or imaginary component.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "bSecondary"
        super().__init__(locations, orientation, component, **kwargs)


class PointMagneticField(BaseRx):
    """Measure FDEM total magnetic field at a point.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation.
    component : {'real', 'imag', 'both', 'complex'}
        Component of the receiver; i.e. 'real' or 'imag'. The options 'both' and
        'complex' are only available for the 1D layered simulations.
    data_type : {'field', 'ppm'}
        Data type observed by the receiver, either field, or ppm secondary
        of the total field.
    use_source_receiver_offset : bool, optional
        Whether to interpret the receiver locations as defining the source and receiver
        offset.

    Notes
    -----
    `data_type`, `use_source_receiver_offset`, and the options of `'both'` and
    `'complex'` for component are only implemented for the `Simulation1DLayered`.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "h"
        super().__init__(locations, orientation, component, **kwargs)


class PointMagneticFieldSecondary(BaseRx):
    """
    Magnetic flux FDEM receiver


    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation
    component : {'real', 'imag', 'both', 'complex'}
        Component of the receiver; i.e. 'real' or 'imag'. The options 'both' and
        'complex' are only available for the 1D layered simulations.
    data_type : {'field', 'ppm'}
        Data type observed by the receiver, either field, or ppm secondary
        of the total field.
    use_source_receiver_offset : bool, optional
        Whether to interpret the receiver locations as defining the source and receiver
        offset.

    Notes
    -----
    `data_type`, `use_source_receiver_offset`, and the options of `'both'` and
    `'complex'` for component are only implemented for the `Simulation1DLayered`.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "hSecondary"
        super().__init__(
            locations, orientation=orientation, component=component, **kwargs
        )


class PointCurrentDensity(BaseRx):
    """Measure FDEM current density at a point.

    Parameters
    ----------
    locations : (n_loc, n_dim) numpy.ndarray
        Receiver locations.
    orientation : {'x', 'y', 'z'}
        Receiver orientation.
    component : {'real', 'imag'}
        Real or imaginary component.
    """

    def __init__(self, locations, orientation="x", component="real", **kwargs):
        self.projField = "j"
        super().__init__(locations, orientation, component, **kwargs)



"""
==============================================================================
Natural Source EM (:mod:`SimPEG.electromagnetics.natural_source`)
==============================================================================
.. currentmodule:: SimPEG.electromagnetics.natural_source

About ``natural_source``

Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation1DElectricField
  Simulation1DMagneticField
  Simulation1DPrimarySecondary
  Simulation2DElectricField
  Simulation2DMagneticField
  Simulation3DPrimarySecondary

Receivers
=========
.. autosummary::
  :toctree: generated/

  receivers.PointNaturalSource
  receivers.Point3DTipper

Sources
=======
.. autosummary::
  :toctree: generated/

  sources.Planewave
  sources.PlanewaveXYPrimary

Data
====
.. autosummary::
  :toctree: generated/

  survey.Data

Fields
======
.. autosummary::
  :toctree: generated/

  Fields1DElectricField
  Fields1DMagneticField
  Fields1DPrimarySecondary
  Fields2DElectricField
  Fields2DMagneticField
  Fields3DPrimarySecondary

"""

from . import utils
from . import sources as Src
from . import receivers as Rx
from .survey import Survey, Data
from .fields import (
    Fields1DElectricField,
    Fields1DMagneticField,
    Fields1DPrimarySecondary,
    Fields2DElectricField,
    Fields2DMagneticField,
    Fields3DPrimarySecondary,
)
from .simulation import (
    Simulation1DElectricField,
    Simulation1DMagneticField,
    Simulation1DPrimarySecondary,
    Simulation2DElectricField,
    Simulation2DMagneticField,
    Simulation3DPrimarySecondary,
)
from . import sources
from . import receivers
from .simulation_1d import Simulation1DRecursive



import numpy as np
from ...potential_fields.base import BasePFSimulation as Sim
import os
from dask import delayed, array, config
from dask.diagnostics import ProgressBar
from ..utils import compute_chunk_sizes

Sim._chunk_format = "row"


@property
def chunk_format(self):
    "Apply memory chunks along rows of G, either 'equal', 'row', or 'auto'"
    return self._chunk_format


@chunk_format.setter
def chunk_format(self, other):
    if other not in ["equal", "row", "auto"]:
        raise ValueError("Chunk format must be 'equal', 'row', or 'auto'")
    self._chunk_format = other


Sim.chunk_format = chunk_format


def dask_dpred(self, m=None, f=None, compute_J=False):
    if m is not None:
        self.model = m
    if f is not None:
        return f
    return self.fields(self.model)


Sim.dpred = dask_dpred


def dask_residual(self, m, dobs, f=None):
    return self.dpred(m, f=f) - dobs


Sim.residual = dask_residual


def dask_linear_operator(self):
    forward_only = self.store_sensitivities is None
    row = delayed(self.evaluate_integral, pure=True)
    n_cells = self.nC
    if getattr(self, "model_type", None) == "vector":
        n_cells *= 3

    rows = [
        array.from_delayed(
            row(receiver_location, components),
            dtype=np.float32,
            shape=(len(components),) if forward_only else (len(components), n_cells),
        )
        for receiver_location, components in self.survey._location_component_iterator()
    ]
    if forward_only:
        stack = array.concatenate(rows)
    else:
        stack = array.vstack(rows)
        # Chunking options
        if self.chunk_format == "row":
            config.set({"array.chunk-size": f"{self.max_chunk_size}MiB"})
            # Autochunking by rows is faster and more memory efficient for
            # very large problems sensitivty and forward calculations
            stack = stack.rechunk({0: "auto", 1: -1})
        elif self.chunk_format == "equal":
            # Manual chunks for equal number of blocks along rows and columns.
            # Optimal for Jvec and Jtvec operations
            row_chunk, col_chunk = compute_chunk_sizes(
                *stack.shape, self.max_chunk_size
            )
            stack = stack.rechunk((row_chunk, col_chunk))
        else:
            # Auto chunking by columns is faster for Inversions
            config.set({"array.chunk-size": f"{self.max_chunk_size}MiB"})
            stack = stack.rechunk({0: -1, 1: "auto"})

    if self.store_sensitivities == "disk":
        sens_name = os.path.join(self.sensitivity_path, "sensitivity.zarr")
        if os.path.exists(sens_name):
            kernel = array.from_zarr(sens_name)
            if np.all(
                np.r_[
                    np.any(np.r_[kernel.chunks[0]] == stack.chunks[0]),
                    np.any(np.r_[kernel.chunks[1]] == stack.chunks[1]),
                    np.r_[kernel.shape] == np.r_[stack.shape],
                ]
            ):
                # Check that loaded kernel matches supplied data and mesh
                print("Zarr file detected with same shape and chunksize ... re-loading")
                return kernel
        else:
            print("Writing Zarr file to disk")
            with ProgressBar():
                print("Saving kernel to zarr: " + sens_name)
                kernel = array.to_zarr(
                    stack, sens_name, compute=True, return_stored=True, overwrite=True
                )
    elif forward_only:
        with ProgressBar():
            print("Forward calculation: ")
            kernel = stack.compute()
    else:
        with ProgressBar():
            print("Computing sensitivities to local ram")
            kernel = stack.persist()
    return kernel


Sim.linear_operator = dask_linear_operator



import numpy as np
import scipy.sparse as sp

from .... import props, maps
from ....data import Data
from ....base import BasePDESimulation

from ..resistivity import Simulation3DCellCentered as DC_3D_CC
from ..resistivity import Simulation3DNodal as DC_3D_N
from ..resistivity import Simulation2DCellCentered as DC_2D_CC
from ..resistivity import Simulation2DNodal as DC_2D_N


class BaseIPSimulation(BasePDESimulation):
    sigma = props.PhysicalProperty("Electrical Conductivity (S/m)")
    rho = props.PhysicalProperty("Electrical Resistivity (Ohm m)")
    props.Reciprocal(sigma, rho)

    @property
    def sigmaMap(self):
        return maps.IdentityMap()

    @sigmaMap.setter
    def sigmaMap(self, arg):
        pass

    @property
    def rhoMap(self):
        return maps.IdentityMap()

    @rhoMap.setter
    def rhoMap(self, arg):
        pass

    @property
    def sigmaDeriv(self):
        return -sp.diags(self.sigma) @ self.etaDeriv

    @property
    def rhoDeriv(self):
        return sp.diags(self.rho) @ self.etaDeriv

    eta, etaMap, etaDeriv = props.Invertible("Electrical Chargeability (V/V)")

    def __init__(
        self,
        mesh=None,
        survey=None,
        sigma=None,
        rho=None,
        eta=None,
        etaMap=None,
        Ainv=None,  # A DC's Ainv
        _f=None,  # A pre-computed DC field
        **kwargs
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.sigma = sigma
        self.rho = rho
        self.eta = eta
        self.etaMap = etaMap
        if Ainv is not None:
            self.Ainv = Ainv
        self._f = _f

    _Jmatrix = None
    _pred = None
    _scale = None

    def fields(self, m):
        if self.verbose:
            print(">> Compute DC fields")
        if self._f is None:
            # re-uses the DC simulation's fields method
            self._f = super().fields(None)

        if self._scale is None:
            scale = Data(self.survey, np.ones(self.survey.nD))
            try:
                f = self.fields_to_space(self._f)
            except AttributeError:
                f = self._f
            # loop through receievers to check if they need to set the _dc_voltage
            for src in self.survey.source_list:
                for rx in src.receiver_list:
                    if rx.data_type == "apparent_chargeability":
                        scale[src, rx] = 1.0 / rx.eval(src, self.mesh, f)
            self._scale = scale.dobs

        self._pred = self.forward(m, f=self._f)

        return self._f

    def dpred(self, m=None, f=None):
        r"""
        Predicted data.

        .. math::

            d_\text{pred} = Pf(m)

        """
        # return self.Jvec(m, m, f=f)
        if f is None:
            f = self.fields(m)

        return self._pred

    def getJtJdiag(self, m, W=None, f=None):
        if getattr(self, "_jtjdiag", None) is None:
            J = self.getJ(m, f=f)
            if W is None:
                W = self._scale**2
            else:
                W = (self._scale * W.diagonal()) ** 2

            self._jtjdiag = np.einsum("i,ij,ij->j", W, J, J)

        return self._jtjdiag

    def Jvec(self, m, v, f=None):
        return self._scale * super().Jvec(m, v, f)

    def forward(self, m, f=None):
        return np.asarray(self.Jvec(m, m, f=f))

    def Jtvec(self, m, v, f=None):
        return super().Jtvec(m, v * self._scale, f)

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = []
        return toDelete


class Simulation2DCellCentered(BaseIPSimulation, DC_2D_CC):
    """
    2.5D cell centered IP problem
    """


class Simulation2DNodal(BaseIPSimulation, DC_2D_N):
    """
    2.5D nodal IP problem
    """


class Simulation3DCellCentered(BaseIPSimulation, DC_3D_CC):
    """
    3D cell centered IP problem
    """


class Simulation3DNodal(BaseIPSimulation, DC_3D_N):
    """
    3D nodal IP problem
    """


Simulation2DCellCentred = Simulation2DCellCentered
Simulation3DCellCentred = Simulation3DCellCentered



"""
========================================================
Utility Classes and Functions (:mod:`SimPEG.utils`)
========================================================
.. currentmodule:: SimPEG.utils

The ``utils`` package contains utilities for helping with common operations involving
SimPEG.

Many of the utilities are imported from `discretize.utils`. See that package's
documentation for many details on items.


Coordinates Utility Functions
=============================

.. autosummary::
  :toctree: generated/

  rotation_matrix_from_normals
  rotate_points_from_normals

Counter Utility Functions
=========================

.. autosummary::
  :toctree: generated/

  Counter
  count
  timeIt

Curvilinear Utility Functions
=============================

.. autosummary::
  :toctree: generated/

  example_curvilinear_grid
  face_info
  index_cube
  volume_tetrahedron


IO Utility Functions
====================

.. autosummary::
  :toctree: generated/

  download
  io_utils.read_dcip2d_ubc
  io_utils.read_dcip3d_ubc
  io_utils.read_dcipoctree_ubc
  io_utils.read_dcip_xyz
  io_utils.read_gg3d_ubc
  io_utils.read_grav3d_ubc
  io_utils.read_mag3d_ubc
  io_utils.write_dcip2d_ubc
  io_utils.write_dcip3d_ubc
  io_utils.write_dcipoctree_ubc
  io_utils.write_dcip_xyz
  io_utils.write_grav3d_ubc
  io_utils.write_gg3d_ubc
  io_utils.write_mag3d_ubc


Matrix Utility Functions
========================

.. autosummary::
  :toctree: generated/

  av
  av_extrap
  cartesian2spherical
  coterminal
  ddx
  define_plane_from_points
  diagEst
  eigenvalue_by_power_iteration
  estimate_diagonal
  get_subarray
  kron3
  ind2sub
  inverse_2x2_block_diagonal
  inverse_3x3_block_diagonal
  inverse_property_tensor
  make_property_tensor
  mkvc
  ndgrid
  sdiag
  sdinv
  speye
  spherical2cartesian
  spzeros
  sub2ind
  unique_rows


Mesh Utility Functions
======================

.. autosummary::
  :toctree: generated/

  closest_points_index
  extract_core_mesh
  unpack_widths
  surface2inds


Model Utility Functions
=======================

.. autosummary::
  :toctree: generated/

  depth_weighting
  surface2ind_topo
  model_builder.add_block
  model_builder.create_2_layer_model
  model_builder.create_block_in_wholespace
  model_builder.create_ellipse_in_wholespace
  model_builder.create_from_function
  model_builder.create_layers_model
  model_builder.create_random_model
  model_builder.get_indices_block
  model_builder.get_indices_polygon
  model_builder.get_indices_sphere


Plotting Utility Functions
==========================

.. autosummary::
  :toctree: generated/

  plot2Ddata
  plot_1d_layer_model


PGI Utility Classes and Functions
=================================
.. autosummary::
  :toctree: generated/

  WeightedGaussianMixture
  GaussianMixtureWithPrior
  GaussianMixtureWithNonlinearRelationships
  GaussianMixtureWithNonlinearRelationshipsWithPrior

Code Utility Functions
======================
Many of the functions here are used internally to SimPEG and have minimal documentation.

.. autosummary::
  :toctree: generated/

  as_array_n_by_dim
  call_hooks
  check_stoppers
  mem_profile_class
  dependent_property
  deprecate_class
  deprecate_function
  deprecate_method
  deprecate_module
  deprecate_property
  hook
  print_done
  printDone
  print_line
  print_stoppers
  print_titles
  requires
  set_kwargs
  validate_float
  validate_integer
  validate_list_of_types
  validate_location_property
  validate_ndarray_with_shape
  validate_string
  validate_callable
  validate_direction
  validate_active_indices

"""
from discretize.utils.interpolation_utils import interpolation_matrix

from .code_utils import (
    mem_profile_class,
    hook,
    set_kwargs,
    print_titles,
    print_line,
    check_stoppers,
    print_stoppers,
    print_done,
    call_hooks,
    deprecate_property,
    deprecate_module,
    deprecate_method,
    deprecate_function,
    deprecate_class,
    dependent_property,
    as_array_n_by_dim,
    requires,
    Report,
    validate_float,
    validate_integer,
    validate_list_of_types,
    validate_location_property,
    validate_ndarray_with_shape,
    validate_string,
    validate_type,
    validate_callable,
    validate_direction,
    validate_active_indices,
)

from .mat_utils import (
    mkvc,
    sdiag,
    sdinv,
    speye,
    kron3,
    spzeros,
    ddx,
    av,
    av_extrap,
    ndgrid,
    ind2sub,
    sub2ind,
    get_subarray,
    inverse_3x3_block_diagonal,
    inverse_2x2_block_diagonal,
    TensorType,
    make_property_tensor,
    inverse_property_tensor,
    estimate_diagonal,
    Zero,
    Identity,
    unique_rows,
    eigenvalue_by_power_iteration,
    cartesian2spherical,
    spherical2cartesian,
    coterminal,
    define_plane_from_points,
)
from .mesh_utils import (
    unpack_widths,
    closest_points_index,
    extract_core_mesh,
    surface2inds,
)
from .curv_utils import (
    volume_tetrahedron,
    index_cube,
    face_info,
    example_curvilinear_grid,
)
from .counter_utils import Counter, count, timeIt
from . import model_builder
from . import solver_utils
from . import io_utils
from .coord_utils import (
    rotation_matrix_from_normals,
    rotate_points_from_normals,
)
from .model_utils import surface2ind_topo, depth_weighting
from .plot_utils import plot2Ddata, plotLayer, plot_1d_layer_model
from .io_utils import download
from .pgi_utils import (
    GaussianMixture,
    WeightedGaussianMixture,
    GaussianMixtureWithPrior,
    GaussianMixtureWithNonlinearRelationships,
    GaussianMixtureWithNonlinearRelationshipsWithPrior,
)

# Deprecated imports
interpmat = deprecate_function(
    interpolation_matrix, "interpmat", removal_version="0.19.0", future_warn=True
)

from .code_utils import (
    memProfileWrapper,
    setKwargs,
    printTitles,
    printLine,
    checkStoppers,
    printStoppers,
    printDone,
    callHooks,
    dependentProperty,
    asArray_N_x_Dim,
)
from .mat_utils import (
    sdInv,
    getSubArray,
    inv3X3BlockDiagonal,
    inv2X2BlockDiagonal,
    makePropertyTensor,
    invPropertyTensor,
    diagEst,
    uniqueRows,
)
from .mesh_utils import (
    meshTensor,
    closestPoints,
    ExtractCoreMesh,
)
from .curv_utils import (
    volTetra,
    faceInfo,
    indexCube,
    exampleLrmGrid,
)
from .coord_utils import (
    rotatePointsFromNormals,
    rotationMatrixFromNormals,
)
